/*
 * Inverter_Atmega_16.c
 *
 * Created: 24/02/2016 18:44:46
 * Author : anders
 */ 
/* General includes */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdbool.h>
#include <stdlib.h>
#define F_CPU 16000000UL
#include <util/delay.h>

/* Specific includes */
#include "adc/adc.h"
#include "timer/timer.h"
#include "sensors/sensor_conversion.h"
#include "atmega64_m1-base/drivers/CAN/can.h"
#include "parameters.h"

/************************************
*
*	Temperature & VDC thresholds
*
************************************/
#define INV_T_ON	9000 //centidegrees C.
#define MOTOR_T_ON	9000
#define VDC_ON		800 
#define INV_T_OFF	2000
#define MOTOR_T_OFF	2000
#define VDC_OFF		650 

#define NUM_CAN_RX_BUFFERS	1
#define ADC_CHANNELS_NUM	2



void callback_can(uint8_t status);
void translateValue(int16_t data, int pos);
void checkForErrors();
void setupIO();

volatile int16_t adc_raw[ADC_CHANNELS_NUM];

// arrays of temperatures for systems, used on the form [T1, T2, T3, TM].
TempData sys_1_temp[4];
TempData sys_2_temp[4];
volatile int16_t vdc_value;

/************************************
*
*		 CAN receive buffers
*
************************************/
can_msg_buffer buffer_open = {
	.id.std = 0xAA,
	.mask.std = 0xFF,
	.callback_function = callback_can
};

can_msg_buffer *rx_info[NUM_CAN_RX_BUFFERS]= {
	&buffer_open
};

void callback_can(uint8_t status){
	can_get_rx_buffer_data(&buffer_open, NULL);
}

/************************************
*
*	CAN messages and their data
*
************************************/ 
can_data_t alive_msg_data;
can_msg_t alive_msg = {
	.id.std = 0x80, //ID of Alive message
	.dlc = 1, // Data length code (number of bytes of data)
	.pt_data = &alive_msg_data
};

can_data_t temperature_inv1_msg_data;
can_msg_t temperature_inv1_msg = {
	.id.std = 0xF1, //ID of Alive message
	.dlc = 8, // Data length code (number of bytes of data)
	.pt_data = &temperature_inv1_msg_data
};

can_data_t temperature_inv2_msg_data;
can_msg_t temperature_inv2_msg = {
	.id.std = 0xF2, //ID of Alive message
	.dlc = 8, // Data length code (number of bytes of data)
	.pt_data = &temperature_inv2_msg_data
};


int main(void){
	setupIO();
	adc_init();
	timer1_init();
	can_init(clkph_16MHZ, CAN_BPS_1000k, rx_info, NUM_CAN_RX_BUFFERS);
	sei();
	
	// Kick off adc continous adc reading. 
	adc_start_continous();

	while (1) {
		for (int i = 0; i<ADC_CHANNELS_NUM; i++){
			translateValue(adc_raw[i], i);
		}
		setErrorPins();
	}
} 

/**********************************************
*
*		Functions
*
**********************************************/
void translateValue(int16_t data, int pos){
	switch(pos){
		case 0: // ADC2, INV2_T3
			sys_2_temp[T3].temp = sensor_adc_to_temp(data,false);
			break;
		case 1: // ADC3, INV2_T2
			sys_2_temp[T2].temp = sensor_adc_to_temp(data,true);
			break;
		case 2: // ADC4, INV1_T3
			sys_1_temp[T3].temp = sensor_adc_to_temp(data,false);
			break;
		case 3: // ADC5, VDC
			vdc_value = sensor_adc_to_vdc(data);
			break;
		case 4: // ADC6, INV1_T1
			sys_1_temp[T1].temp = sensor_adc_to_temp(data,false);
			break;
		case 5: // ADC7, INV1_T2
			sys_1_temp[T2].temp = sensor_adc_to_temp(data,false);
			break;
		case 6: // ADC8, INV1_TM
			sys_1_temp[TM].temp = sensor_adc_to_temp(data,true);
			break;
		case 7: //9 ADC9, INV2_TM
			sys_2_temp[TM].temp = sensor_adc_to_temp(data,true);
			break;
		case 8: // ADC10, INV2_T1
			sys_2_temp[T1].temp = sensor_adc_to_temp(data, false);
			break;
	}
}

void setErrorPins(){
	
	/** System 1 temperature check **/
	for (int i = 0; i<3; i++){
		sensor_hysteresis(INV_T_ON, INV_T_OFF, &sys_1_temp[i]);
	}
	sensor_hysteresis(MOTOR_T_ON, MOTOR_T_OFF, &sys_1_temp[TM]);
	
	if(sys_1_temp[T1].hysteresis | sys_1_temp[T2].hysteresis | sys_1_temp[T3].hysteresis |sys_1_temp[TM].hysteresis){
		PORTC |= (1 << E_T1_PIN);
	}else{
		PORTC &= ~(1 << E_T1_PIN);
	}
	
	/** System 2 temperature check **/
	for (int i = 0; i<3; i++){
		sensor_hysteresis(INV_T_ON, INV_T_OFF, &sys_2_temp[i]);
	}
	sensor_hysteresis(MOTOR_T_ON, MOTOR_T_OFF, &sys_2_temp[TM]);
	
	if(sys_2_temp[T1].hysteresis | sys_2_temp[T2].hysteresis | sys_2_temp[T3].hysteresis |sys_2_temp[TM].hysteresis){
		PORTC |= (1 << E_T2_PIN);
	}else{
		PORTC &= ~(1 << E_T2_PIN);
	}
	if(sys_2_temp[T3].hysteresis){ 
		PORTD |= (1 << E_G2_PIN);
	}else{
		PORTD &= ~(1 << E_G2_PIN);
	}	
	
	if(sys_2_temp[TM].hysteresis){
		PORTD |= (1 << E_G1_PIN);
	}else{
		PORTD &= ~(1 << E_G1_PIN);
	}
	
	/** Vdc check **/
	static bool e_g_on;
	if(!e_g_on){ // error pin is off
		if(vdc_value > VDC_ON){ // have high voltage
			PORTD |= (1<<E_G1_PIN) | (1<<E_G2_PIN);
			e_g_on = true;
		}
	}else{
		if(vdc_value < VDC_OFF){
			PORTD &= ~(1<<E_G1_PIN) | ~(1<<E_G2_PIN);
			e_g_on = false;
		}
	}
	
	/** Overcurrent pin check **/
	if(!(PINC & (1 << OCA1_PIN)) || !(PINB & (1 << OCB1_PIN))){
		PORTB |= (1 << E_C1_PIN);
	}else{
		PORTB &= ~(1 << E_C1_PIN);
	}
	if(!(PINB & (1 << OCA2_PIN)) || !(PINB & (1 << OCB2_PIN))){
		PORTD |= (1 << E_C2_PIN);
	}else{
		PORTD &= ~(1 << E_C2_PIN);
	}
}

void setupIO(){	
	// Set output ports
	DDRB |= (1 << E_C1_PIN);
	DDRC |= (1 << E_T1_PIN) | (1 << E_T2_PIN);
	DDRD |= (1 << E_G1_PIN) | (1 << E_G2_PIN) | (1 << E_C2_PIN);
	
	// Set internal pull-up
	PORTB |= (1 << OCA2_PIN) | (1 << OCB1_PIN) | (1 << OCB2_PIN);
	PORTC |= (1 << OCA1_PIN);
}
/**********************************************
*
*		Interrupt service routines
*
**********************************************/

ISR(ADC_vect){
	static int8_t adc_index = 0;
	
	// Read ADC conversion result
	adc_raw[adc_index] = ADC;
	
	// Increment channel to be read
	adc_index ++;
	adc_index = adc_index % ADC_CHANNELS_NUM;
	adc_set_source(adc_index + 2); // +2: want to read ADC2..10
	
	// Start new conversion
	ADCSRA |= (1<<ADSC);
	_delay_us(10);
}

ISR(TIMER1_COMPA_vect){
	// Timer 1 @ 10Hz
	// Update message data
	for(int i = 0; i<4; i++){
		temperature_inv1_msg_data.i16[i] = sys_1_temp[i].temp;
		temperature_inv2_msg_data.i16[i] = sys_2_temp[i].temp;
	}
	
	
	/***** MESSAGE TESTING ****/
	/*temperature_inv2_msg_data.i16[0] = adc_raw[0]; // test switch temp formula
	temperature_inv2_msg_data.i16[1] = sys_2_temp[T3].temp;
	temperature_inv2_msg_data.i16[2] = adc_raw[1];
	temperature_inv2_msg_data.i16[3] = sys_2_temp[TM].temp;*/
	can_send_message(&temperature_inv1_msg);
	_delay_us(1000);
	can_send_message(&temperature_inv2_msg);
	_delay_us(1000);
	
	static int8_t alive_timer = 0;
	// Executes @1 Hz. 
	if(alive_timer == 9){
		alive_msg_data.u8[0] = 0x00;
		can_send_message(&alive_msg);
		alive_timer = 0;
	}
	alive_timer ++;
}
