/**
 * \file
 *
 * \brief Synchronous Serial Controller (SSC) driver for SAM.
 *
 * Copyright (c) 2011-2015 Atmel Corporation. All rights reserved.
 *
 * \asf_license_start
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 *    Atmel microcontroller product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * \asf_license_stop
 *
 */
/*
 * Support and FAQ: visit <a href="http://www.atmel.com/design-support/">Atmel Support</a>
 */

#include <sam.h>
#include "../Sam-base/RevolveDrivers/pio.h"
#include "../Sam-base/RevolveDrivers/pmc.h"
#include <string.h>
#include "ssc.h"

void ssc_set_receiver(Ssc *p_ssc, Ssc_opt_t *p_ssc_opt);
void ssc_set_transmitter(Ssc *p_ssc, Ssc_opt_t *p_ssc_opt);
void ssc_enable_interrupt(Ssc *p_ssc, uint32_t ul_sources);
void ssc_enable_rx(Ssc *p_ssc);
void ssc_enable_tx(Ssc *p_ssc);

/**
 * \brief Set up clock.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param ul_bitrate Desired bit clock.
 * \param ul_mck MCK clock.
 *
 * \retval SSC_RC_YES Success.
 * \retval SSC_RC_NO Invalid input value.
 */
uint32_t ssc_set_clock_divider(Ssc *p_ssc, uint32_t bitrate)
{
	if (bitrate) {
		p_ssc->SSC_CMR = SSC_CMR_DIV(((pmc_getPeripheralClockValue() + bitrate) / bitrate) >> 1);
		return SSC_RC_YES;
	} else {
		return SSC_RC_NO;
	}
}

void ssc_init(Ssc *p_ssc, Ssc_opt_t *opts){
	pmc_enable_periph_clk(SSC_IRQn);
	
	ssc_set_clock_divider(p_ssc, opts->baud_rate);
	
	/* Configure the SSC transmitter. */
	ssc_set_transmitter(p_ssc, opts);
	ssc_set_receiver(p_ssc, opts);
		
	ssc_enable_interrupt(p_ssc, opts->interrupt_mask);
	
	ssc_enable_tx(p_ssc);
	ssc_enable_rx(p_ssc);
}

/**
 * \brief Reset SSC module.
 *
 * \param p_ssc Pointer to an SSC instance.
 */
void ssc_reset(Ssc *p_ssc)
{
	p_ssc->SSC_CR = SSC_CR_SWRST;
	p_ssc->SSC_CMR = 0;
	p_ssc->SSC_RCMR = 0;
	p_ssc->SSC_RFMR = 0;
	p_ssc->SSC_TCMR = 0;
	p_ssc->SSC_TFMR = 0;
}

/**
 * \brief Enable SSC receiver.
 *
 * \param p_ssc Pointer to an SSC instance.
 */
void ssc_enable_rx(Ssc *p_ssc)
{
	p_ssc->SSC_CR = SSC_CR_RXEN;
}

/**
 * \brief Disable SSC receiver.
 *
 * \param p_ssc Pointer to an SSC instance.
 */
void ssc_disable_rx(Ssc *p_ssc)
{
	p_ssc->SSC_CR = SSC_CR_RXDIS;
}

/**
 * \brief Enable SSC Transmitter.
 *
 * \param p_ssc Pointer to an SSC instance.
 */
void ssc_enable_tx(Ssc *p_ssc)
{
	p_ssc->SSC_CR = SSC_CR_TXEN;
}

/**
 * \brief Disable SSC Transmitter.
 *
 * \param p_ssc Pointer to an SSC instance.
 */
void ssc_disable_tx(Ssc *p_ssc)
{
	p_ssc->SSC_CR = SSC_CR_TXDIS;
}

/**
 * \brief Configure SSC to work in normal mode.
 *
 * \param p_ssc Pointer to an SSC instance.
 */
void ssc_set_normal_mode(Ssc *p_ssc)
{
	p_ssc->SSC_RFMR &= ~SSC_RFMR_LOOP;
}

/**
 * \brief Configure SSC to work in loop mode.
 *
 * \param p_ssc Pointer to an SSC instance.
 */
void ssc_set_loop_mode(Ssc *p_ssc)
{
	p_ssc->SSC_RFMR |= SSC_RFMR_LOOP;
}

/**
* \brief Configure SSC receive stop selection.
*
* \param p_ssc Pointer to an SSC instance.
* \param ul_sel Compare 0 used or Compare both 0 & 1 used.
*/
/*
void ssc_set_rx_stop_selection(Ssc *p_ssc, uint32_t ul_sel)
{
	if (SSC_RX_STOP_COMPARE_0_1 == ul_sel) {
		p_ssc->SSC_RCMR |= SSC_RCMR_STOP;
		} else if (SSC_RX_STOP_COMPARE_0 == ul_sel) {
		p_ssc->SSC_RCMR &= ~SSC_RCMR_STOP;
	}
}*/

/**
 * \brief Configure SSC default level driven on the TD pin while
 * out of transmission.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param ul_level The default driven level of TD pin.
 */
void ssc_set_td_default_level(Ssc *p_ssc, uint32_t ul_level)
{
	if (ul_level) {
		p_ssc->SSC_TFMR |= SSC_TFMR_DATDEF;
	} else {
		p_ssc->SSC_TFMR &= ~SSC_TFMR_DATDEF;
	}
}

/**>
 * \brief The TD line is driven with the SSC_TSHR register value
 * during the transmission of the Transmit Frame Sync Signal.
 *
 * \param p_ssc Pointer to an SSC instance.
 */
void ssc_enable_tx_frame_sync_data(Ssc *p_ssc)
{
	p_ssc->SSC_TFMR |= SSC_TFMR_FSDEN;
}

/**
 * \brief The TD line is driven with the default value during the Transmit
 * Frame Sync signal.
 *
 * \param p_ssc Pointer to an SSC instance.
 */
void ssc_disable_tx_frame_sync_data(Ssc *p_ssc)
{
	p_ssc->SSC_TFMR &= ~SSC_TFMR_FSDEN;
}

/**
 * \brief Configure SSC receiver clock mode and date frame configuration.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param p_rx_clk_opt Pointer to the receiver clock configuration structure.
 * \param p_rx_data_frame Pointer to the receiver data frame configuration structure.
 */
void ssc_set_receiver(Ssc *p_ssc, Ssc_opt_t *p_ssc_opt)
{
	if (p_ssc_opt == NULL) {
		p_ssc->SSC_RCMR = 0;
	} else {
		p_ssc->SSC_RCMR |=	SSC_RCMR_CKS(p_ssc_opt->rx_opts.clock_source) |
							SSC_RCMR_CKO(p_ssc_opt->rx_opts.clock_output_mode) |
							SSC_RCMR_CKG(p_ssc_opt->rx_opts.clock_gating_mode)|
							SSC_RCMR_START(p_ssc_opt->rx_opts.communication_trigger) |
							SSC_RCMR_PERIOD(p_ssc_opt->tx_opts.period) |
							SSC_RCMR_STTDLY(p_ssc_opt->rx_opts.delay_ticks);
		if(p_ssc_opt->rx_opts.sample_data_on_rising_clk){
			p_ssc->SSC_RCMR |= SSC_RCMR_CKI;
		}
	}

	uint32_t fl_ext = 0;
	uint32_t fl = 0;
	if(p_ssc_opt->rx_opts.frame_sync_length > 15){
		fl_ext = p_ssc_opt->rx_opts.frame_sync_length / 16;
	}
	fl = p_ssc_opt->rx_opts.frame_sync_length % 16 - 1;
	p_ssc->SSC_RFMR |=	SSC_TFMR_DATLEN(p_ssc_opt->rx_opts.data_length-1) |
						p_ssc_opt->rx_opts.bit_dir |
						SSC_RFMR_DATNB(p_ssc_opt->rx_opts.number_of_data-1) |
						SSC_RFMR_FSLEN(fl) |
						SSC_RFMR_FSLEN_EXT(fl_ext) |
						SSC_RFMR_FSOS(p_ssc_opt->rx_opts.framesync_output_polarity )|
						( p_ssc_opt->rx_opts.framesync_interrupt_polarity - 1 ) << 24;
}

/**
 * \brief Configure SSC transmitter clock mode and date frame configuration.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param p_tx_clk_opt Pointer to the transmitter clock configuration structure.
 * \param p_tx_data_frame Pointer to the transmitter data frame configuration structure.
 */
void ssc_set_transmitter(Ssc *p_ssc, Ssc_opt_t *p_ssc_opts)
{
	if (p_ssc_opts == NULL) {
		p_ssc->SSC_TCMR = 0;
	} else {
		p_ssc->SSC_TCMR |= SSC_TCMR_CKS(p_ssc_opts->tx_opts.clock_source) |
				SSC_TCMR_CKO(p_ssc_opts->tx_opts.clock_output_mode) | 
				SSC_TCMR_CKG(p_ssc_opts->tx_opts.clock_gating_mode)|
				SSC_TCMR_START(p_ssc_opts->tx_opts.communication_trigger) |
				SSC_TCMR_PERIOD(p_ssc_opts->tx_opts.period) |
				SSC_TCMR_STTDLY(p_ssc_opts->tx_opts.delay_ticks);
		if(p_ssc_opts->rx_opts.sample_data_on_rising_clk){
			p_ssc->SSC_RCMR |= SSC_RCMR_CKI;
		}
	}

	uint32_t fl_ext = 0;
	uint32_t fl = 0;
	if(p_ssc_opts->tx_opts.frame_sync_length > 15){
		fl_ext = p_ssc_opts->tx_opts.frame_sync_length / 16;
	}
	fl = p_ssc_opts->tx_opts.frame_sync_length % 16 - 1;
	p_ssc->SSC_TFMR |= SSC_TFMR_DATLEN(p_ssc_opts->tx_opts.data_length-1) |
			p_ssc_opts->tx_opts.bit_dir |
			SSC_TFMR_DATNB(p_ssc_opts->tx_opts.number_of_data-1) |
			SSC_TFMR_FSLEN(fl) |
			SSC_TFMR_FSLEN_EXT(fl_ext) |
			SSC_TFMR_FSOS(p_ssc_opts->tx_opts.framesync_output_polarity )|
			( p_ssc_opts->tx_opts.framesync_interrupt_polarity - 1 ) << 24;
}

/**
 * \brief Configure SSC Receive Compare Register.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param ul_id Compare register ID.
 * \param ul_value Value to configure.
 */
void ssc_set_rx_compare(Ssc *p_ssc, uint32_t ul_id, uint32_t ul_value)
{
	switch (ul_id) {
	case 0:
		p_ssc->SSC_RC0R = ul_value;
		break;
	case 1:
		p_ssc->SSC_RC1R = ul_value;
		break;
	}
}

/**
 * \brief Get SSC Receive Compare Register.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param ul_id Compare register ID.
 *
 * \return Receive Compare Register value for the specified ul_id, otherwise SSC_RC_INVALID.
 */
uint32_t ssc_get_rx_compare(Ssc *p_ssc, uint32_t ul_id)
{
	switch (ul_id) {
	case 0:
		return p_ssc->SSC_RC0R;
	case 1:
		return p_ssc->SSC_RC1R;
	default:
		return SSC_RC_INVALID;
	}
}

/**
 * \brief Enable SSC interrupts.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param ul_sources Interrupts to be enabled.
 */
void ssc_enable_interrupt(Ssc *p_ssc, uint32_t ul_sources)
{
	p_ssc->SSC_IER = ul_sources;
}

/**
 * \brief Disable SSC interrupts.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param ul_sources Interrupts to be enabled.
 */
void ssc_disable_interrupt(Ssc *p_ssc, uint32_t ul_sources)
{
	p_ssc->SSC_IDR = ul_sources;
}

/**
 * \brief Read SSC interrupt mask.
 *
 * \param p_ssc Pointer to an SSC instance.
 *
 * \return The interrupt mask value.
 */
uint32_t ssc_get_interrupt_mask(Ssc *p_ssc)
{
	return p_ssc->SSC_IMR;
}

/**
 * \brief Read SSC status.
 *
 * \param p_ssc Pointer to an SSC instance.
 *
 * \return The SSC status value.
 */
uint32_t ssc_get_status(Ssc *p_ssc)
{
	return p_ssc->SSC_SR;
}

/**
 * \brief Check if data has been loaded in SSC_THR and is waiting to be loaded
 * in the Transmit Shift Register (TSR).
 *
 * \param p_ssc Pointer to an SSC instance.
 *
 * \retval SSC_RC_YES There is no data in the SSC_THR.
 * \retval SSC_RC_NO There is one data in the SSC_THR.
 */
uint32_t ssc_is_tx_ready(Ssc *p_ssc)
{
	if (p_ssc->SSC_SR & SSC_SR_TXRDY) {
		return SSC_RC_YES;
	}
	return SSC_RC_NO;
}

/**
 * \brief Check if the last data written in SSC_THR has been loaded in TSR
 * and the last data loaded in TSR has been transmitted.
 *
 * \param p_ssc Pointer to an SSC instance.
 *
 * \retval SSC_RC_YES Both of the two registers are empty.
 * \retval SSC_RC_NO At least one of the two registers is not empty.
 */
uint32_t ssc_is_tx_empty(Ssc *p_ssc)
{
	if (p_ssc->SSC_SR & SSC_SR_TXEMPTY) {
		return SSC_RC_YES;
	}
	return SSC_RC_NO;
}

/**
 * \brief Check if data has been received and loaded in SSC_RHR.
 *
 * \param p_ssc Pointer to an SSC instance.
 *
 * \retval SSC_RC_YES There is one data in the SSC_RHR.
 * \retval SSC_RC_NO There is no data in the SSC_RHR.
 */
uint32_t ssc_is_rx_ready(Ssc *p_ssc)
{
	if (p_ssc->SSC_SR & SSC_SR_RXRDY) {
		return SSC_RC_YES;
	}
	return SSC_RC_NO;
}

/**
 * \brief Check if transmitter is enabled.
 *
 * \param p_ssc Pointer to an SSC instance.
 *
 * \retval SSC_RC_YES The transmitter is enabled.
 * \retval SSC_RC_NO The transmitter is disabled.
 */
uint32_t ssc_is_tx_enabled(Ssc *p_ssc)
{
	if (p_ssc->SSC_SR & SSC_SR_TXEN) {
		return SSC_RC_YES;
	}
	return SSC_RC_NO;
}

/**
 * \brief Check if receiver is enabled.
 *
 * \param p_ssc Pointer to an SSC instance.
 *
 * \retval SSC_RC_YES The receiver is enabled.
 * \retval SSC_RC_NO The receiver is disabled.
 */
uint32_t ssc_is_rx_enabled(Ssc *p_ssc)
{
	if (p_ssc->SSC_SR & SSC_SR_RXEN) {
		return SSC_RC_YES;
	}
	return SSC_RC_NO;
}

/**
 * \brief Write to SSC Transmit Holding Register.
 * Send data through SSC Data frame.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param ul_frame Frame data to be transmitted.
 *
 * \retval SSC_RC_ERROR Time-out.
 * \retval SSC_RC_OK Success.
 *
 */
uint32_t ssc_write(Ssc *p_ssc, uint32_t ul_frame)
{

	while (!(p_ssc->SSC_SR & SSC_SR_TXEMPTY)) {	}

	p_ssc->SSC_THR = ul_frame;
	return SSC_RC_OK;
}

/**
 * \brief Read from SSC Receive Holding Register.
 * Read data that is received in SSC Data frame.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param ul_data Pointer to the location where to store the received data.
 *
 * \retval SSC_RC_ERROR Time-out.
 * \retval SSC_RC_OK Success.
 */
uint32_t ssc_read(Ssc *p_ssc, uint32_t *ul_data)
{
	while (!(p_ssc->SSC_SR & SSC_SR_RXRDY)) {
	}

	*ul_data = p_ssc->SSC_RHR;
	return SSC_RC_OK;
}

/**
 * \brief Write to SSC Transmit Synchronization Holding Register.
 * Send data through SSC Synchronization frame. If there is sync data that needs to be
 * transmitted, call this function first to send out the sync data, and then call the
 * ssc_write() function to send out application data.
 *
 * \param p_ssc Pointer to an SSC instance.
 * \param ul_frame Frame Synchronization data.
 */
void ssc_write_sync_data(Ssc *p_ssc, uint32_t ul_frame)
{
	p_ssc->SSC_TSHR = ul_frame;
}

/**
 * \brief Read from SSC Receive Synchronization Holding Register.
 * Read data that is received in SSC Synchronization frame. When the sync data is actually
 * used, after successfully reading the application data by calling ssc_read(), call
 * this function, and the return sync data is useful.
 *
 * \param p_ssc Pointer to an SSC instance.
 *
 * \return Current RSHR value.
 */
uint32_t ssc_read_sync_data(Ssc *p_ssc)
{
	return p_ssc->SSC_RSHR;
}