/**
 * \file
 *
 * \brief Synchronous Serial Controller (SSC) driver for SAM.
 *
 * Copyright (c) 2011-2015 Atmel Corporation. All rights reserved.
 *
 * \asf_license_start
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 *    Atmel microcontroller product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * \asf_license_stop
 *
 */
/*
 * Support and FAQ: visit <a href="http://www.atmel.com/design-support/">Atmel Support</a>
 */

#ifndef SSC_H_INCLUDED
#define SSC_H_INCLUDED

#include <sam.h>

//! \brief SSC driver return codes.
enum ssc_return_code {
	SSC_RC_OK = 0,              //!< OK
	SSC_RC_YES = 0,             //!< Yes
	SSC_RC_NO = 1,              //!< No
	SSC_RC_ERROR = 1,           //!< General error
	SSC_RC_INVALID = 0xFFFFFFFF //!< Parameter invalid
};

typedef enum {
	SSC_LSB_FIRST = 0,
	SSC_MSB_FIRST = SSC_TFMR_MSBF
}Ssc_bit_dir_t;

typedef enum{
	SSC_FRAMESYNC_NONE		,
	SSC_FRAMESYNC_FALLING	,
	SSC_FRAMESYNC_RISING	,
	SSC_FRAMESYNC_LOW		,
	SSC_FRAMESYNC_HIGH		,
	SSC_FRAMESYNC_CHANGE	
}Ssc_framesync_polarity;

typedef enum {
	SSC_CLKIN_MCK,
	SSC_CLKIN_XK_CLK,
	SSC_CLKIN_XK_PIN
}Ssc_CLK_in_t;

typedef enum {
	SSC_CLKOUT_NONE,
	SSC_CLKOUT_CONTINUOUS,
	SSC_CLKOUT_TRANSFER
}Ssc_CLK_output_t;

typedef enum {
	SSC_CLKGATE_CONTINUOUS,
	SSC_CLKGATE_XF_LOW,
	SSC_CLKGATE_XF_HIGH
}Ssc_CLK_gating_t;

typedef enum {
	SSC_COMTRIGGER_CONTINUOUS,
	SSC_COMTRIGGER_XMIT,
	SSC_COMTRIGGER_XF_LOW,
	SSC_COMTRIGGER_XF_HIGH,
	SSC_COMTRIGGER_XF_FALLING,
	SSC_COMTRIGGER_XF_RISING,
	SSC_COMTRIGGER_XF_LEVEL_CHANGE,
	SSC_COMTRIGGER_XF_EDGE,
	SSC_COMTRIGGER_CMP0
}Ssc_comm_trigger_t;

//! Data frame structure.
typedef struct {
	uint32_t data_length;
	Ssc_bit_dir_t bit_dir;
	uint32_t number_of_data;
	uint32_t frame_sync_length;
	Ssc_framesync_polarity framesync_output_polarity;
	Ssc_framesync_polarity framesync_interrupt_polarity;
	Ssc_CLK_gating_t clock_gating_mode;
	Ssc_comm_trigger_t communication_trigger;
	Ssc_CLK_output_t clock_output_mode;
	Ssc_CLK_in_t clock_source;
	bool sample_data_on_rising_clk;
	uint32_t delay_ticks;
	uint32_t period;
	uint32_t compare_val_0;
	uint32_t compare_val_1;
	uint32_t default_state;
} Ssc_data_frame_opt_t;

//! Clock mode structure.
typedef struct {
	Ssc_data_frame_opt_t tx_opts;
	Ssc_data_frame_opt_t rx_opts;
	uint32_t baud_rate;
	uint32_t interrupt_mask;
} Ssc_opt_t;

void ssc_init(Ssc *p_ssc, Ssc_opt_t *opts);
void ssc_enable_rx(Ssc *p_ssc);
void ssc_disable_rx(Ssc *p_ssc);
uint32_t ssc_write(Ssc *p_ssc, uint32_t ul_frame);
uint32_t ssc_read(Ssc *p_ssc, uint32_t *ul_data);

#endif /* SSC_H_INCLUDED */
