/*
 * foc.c
 *
 * Created: 01.03.2016 17.01.15
 *  Author: Simen
 */ 

#include <sam.h>
#include <arm_math.h>
#include "../system_state.h"
#include "../utilities.h"
#include "motor.h"

float32_t foc_torque_to_Iq(float32_t torque){
	Motor_data_t *motor_ptr = motor_get_parameter_ptr();
	return (2 * torque) / (3 * motor_ptr->poles * motor_ptr->lambda);
}

float32_t foc_calc_d_current(float32_t current_q){
	Motor_data_t *motor_ptr = motor_get_parameter_ptr();
	float32_t res1 = motor_ptr->lambda / (2 * (motor_ptr->L_q - motor_ptr->L_d));
	float32_t res2 = square(res1) + square(current_q);
	float32_t res3;
	
	arm_sqrt_f32(res2, &res3);
	
	return res1 - res3;
}

void foc_limit_vector_dq(float32_t* d, float32_t* q, float32_t limit){
	if(get_squared_vector_magnitude(*d, *q) > square(limit)){
		float32_t q_max = 0;
		arm_sqrt_f32(square(limit) - square(*d), &q_max);
		*q = copysignf(*q, q_max);
	}
}

void foc_calc_setpoints(state_variables_t *data){
	
	Motor_data_t *motor_ptr = motor_get_parameter_ptr();
	
	data->setpoint_q = foc_torque_to_Iq(data->torque_request);
	
	float32_t last_voltage_magnitude = get_squared_vector_magnitude(data->regulator_data.last_voltage_d, data->regulator_data.last_voltage_q);
	
	data->regulator_data.voltage_saturation = (last_voltage_magnitude >= square(data->voltage_dc));
	
	if(data->regulator_data.voltage_saturation){
		//Field weakening, current controllers are saturated
		data->regulator_data.use_mtpa = false;
		data->setpoint_d -= data->regulator_data.field_weakening_gain_down * (motor_ptr->speed_max - data->speed);
		if(data->setpoint_d < motor_ptr->Id_min){
			data->setpoint_d = motor_ptr->Id_min;
		}
	}
	else if(data->regulator_data.use_mtpa){
		//Linear torque area
		data->current_d = foc_calc_d_current(data->setpoint_q);
	}
	else{
		//Field weakening, current controllers are not saturated
		float32_t current_d_mtpa = foc_calc_d_current(data->setpoint_q);
		data->setpoint_d += data->regulator_data.field_weakening_gain_up * (motor_ptr->speed_max - data->speed);
		
		if(data->setpoint_d < current_d_mtpa){
			data->setpoint_d = current_d_mtpa;
			data->regulator_data.use_mtpa = true;
		}
	}
	
	//Make sure that we are not applying too much D-current
	if(data->setpoint_d > motor_ptr->Id_max){
		data->setpoint_d = motor_ptr->Id_max;
	}
	
	foc_limit_vector_dq(&(data->setpoint_d), &(data->setpoint_q), motor_ptr->I_max);
}

void foc_do_dq_transform(state_variables_t *state){
	float32_t alpha;
	float32_t beta;
	
	arm_clarke_f32(state->current_a, state->current_b, &alpha, &beta);
	arm_park_f32(alpha, beta, &(state->current_d), &(state->current_q), state->sin_val, state->cos_val);
	
}

void foc_do_inverse_dq_transform(state_variables_t *state){
	float32_t alpha;
	float32_t beta;
		
	arm_inv_park_f32(state->voltage_d, state->voltage_q, &alpha, &beta, state->sin_val, state->cos_val);
	arm_inv_clarke_f32(alpha, beta, &(state->voltage_a), &(state->voltage_b));
	
}