/**
 * \file
 *
 * \brief Analog-Front-End Controller driver for SAM.
 *
 * Copyright (c) 2014-2015 Atmel Corporation. All rights reserved.
 *
 * \asf_license_start
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 *    Atmel microcontroller product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * \asf_license_stop
 *
 */
/*
 * Support and FAQ: visit <a href="http://www.atmel.com/design-support/">Atmel Support</a>
 */

#include <sam.h>
#include <stdlib.h>
#include "adc.h"
#include "../pmc.h"


/**
 * \defgroup sam_drivers_adc_group Analog-Front-End Controller
 *
 * See \ref sam_adc_quickstart.
 *
 * Driver for the Analog-Front-End Controller. This driver provides access to
 * the main features of the AFEC controller.
 *
 * @{
 */
/** AFEC interrupt source type */
enum adc_interrupt_source {
	AFEC_INTERRUPT_EOC_0 = 0,
	AFEC_INTERRUPT_EOC_1,
	AFEC_INTERRUPT_EOC_2,
	AFEC_INTERRUPT_EOC_3,
	AFEC_INTERRUPT_EOC_4,
	AFEC_INTERRUPT_EOC_5,
	AFEC_INTERRUPT_EOC_6,
	AFEC_INTERRUPT_EOC_7,
	AFEC_INTERRUPT_EOC_8,
	AFEC_INTERRUPT_EOC_9,
	AFEC_INTERRUPT_EOC_10,
	AFEC_INTERRUPT_EOC_11,
	AFEC_INTERRUPT_DATA_READY,
	AFEC_INTERRUPT_OVERRUN_ERROR,
	AFEC_INTERRUPT_COMP_ERROR,
	AFEC_INTERRUPT_TEMP_CHANGE,
	_AFEC_NUM_OF_INTERRUPT_SOURCE,
	AFEC_INTERRUPT_ALL = 0x47000FFF,
};

/** Definitions for AFEC Start Up Time */
enum adc_startup_time {
	AFEC_STARTUP_TIME_0 = AFEC_MR_STARTUP_SUT0,
	AFEC_STARTUP_TIME_1 = AFEC_MR_STARTUP_SUT8,
	AFEC_STARTUP_TIME_2 = AFEC_MR_STARTUP_SUT16,
	AFEC_STARTUP_TIME_3 = AFEC_MR_STARTUP_SUT24,
	AFEC_STARTUP_TIME_4 = AFEC_MR_STARTUP_SUT64,
	AFEC_STARTUP_TIME_5 = AFEC_MR_STARTUP_SUT80,
	AFEC_STARTUP_TIME_6 = AFEC_MR_STARTUP_SUT96,
	AFEC_STARTUP_TIME_7 = AFEC_MR_STARTUP_SUT112,
	AFEC_STARTUP_TIME_8 = AFEC_MR_STARTUP_SUT512,
	AFEC_STARTUP_TIME_9 = AFEC_MR_STARTUP_SUT576,
	AFEC_STARTUP_TIME_10 = AFEC_MR_STARTUP_SUT640,
	AFEC_STARTUP_TIME_11 = AFEC_MR_STARTUP_SUT704,
	AFEC_STARTUP_TIME_12 = AFEC_MR_STARTUP_SUT768,
	AFEC_STARTUP_TIME_13 = AFEC_MR_STARTUP_SUT832,
	AFEC_STARTUP_TIME_14 = AFEC_MR_STARTUP_SUT896,
	AFEC_STARTUP_TIME_15 = AFEC_MR_STARTUP_SUT960
};

/** Definitions for AFEC power mode */
enum adc_power_mode {
	/* AFEC core on and reference voltage circuitry on */
	AFEC_POWER_MODE_ALL_ON = 0,
	/* AFEC core off and reference voltage circuitry on */
	AFEC_POWER_MODE_REF_ON,
	/* AFEC core off and reference voltage circuitry off */
	AFEC_POWER_MODE_ALL_OFF
};

#define AFEC_CLK_HZ	40000000
#define AFEC_TUNE_OFFSET_TARGET 140
#define AFEC_MAX_OFFSET 1023
#define AFEC_CGR_set(channel, gain)	(gain << (channel*2))
#define AFEC_CGR_msk(channel)		AFEC_CGR_set(channel, 0x3u)


uint16_t adc_tune_offset(Afec *afec, adc_channel_num_t channel);
void adc_set_power_mode(Afec *const afec, const enum adc_power_mode mode);
static void adc_channel_set_analog_offset(Afec *const afec, adc_channel_num_t adc_ch, uint16_t aoffset);
static void adc_channel_enable(Afec *const afec, adc_channel_num_t adc_ch);
void adc_set_conversion_trigger(Afec *const afec, adc_trigger_t conversion_trigger);
void adc_set_channel_gain(Afec *const afec, adc_channel_num_t channel, adc_gain_t gain);
/**
 * \brief Initialize the AFEC Module.
 *
 * \param afec  Base address of the AFEC
 * \param config   Configuration for the AFEC
 *
 * \retval STATUS_OK  Initialization is finished.
 * \retval STATUS_ERR_BUSY  Initialization failed.
 */
uint32_t adc_init(Afec *const afec, adc_opts_t *config, adc_channel_opts_t enabled_channels[], uint8_t num_channels)
{
	uint32_t reg;
	
	if(afec == AFEC0){
		pmc_enable_periph_clk(AFEC0_IRQn);
	}
	if(afec == AFEC1){
		pmc_enable_periph_clk(AFEC1_IRQn);
	}
	
	if ((afec->AFEC_ISR & AFEC_ISR_DRDY) == AFEC_ISR_DRDY) {
		return 1;
	}

	/* Reset and configure the AFEC module */
	afec->AFEC_CR = AFEC_CR_SWRST;
		
	reg =	AFEC_MR_PRESCAL((pmc_getPeripheralClockValue() / AFEC_CLK_HZ )- 1) |
			AFEC_MR_ONE |
			AFEC_MR_TRACKTIM(15) |
			AFEC_MR_TRANSFER(3) |
			AFEC_MR_STARTUP_SUT64;

	afec->AFEC_MR = reg;
	
	reg = 0;
	
	reg |= config->resolution | AFEC_EMR_STM  | AFEC_EMR_TAG;
	
	afec->AFEC_EMR = reg;
	
	afec->AFEC_EMR |= config->sign_mode;
	
	adc_set_power_mode(afec, AFEC_POWER_MODE_ALL_ON);
	
	adc_set_conversion_trigger(afec, config->conversion_trigger);
	
	afec->AFEC_ACR = AFEC_ACR_PGA0EN | AFEC_ACR_PGA1EN;
	
	for(uint8_t i = 0; i < num_channels; i++){
		adc_channel_opts_t *channel_pt = enabled_channels + i;
		if(channel_pt->differential_channel){
			afec->AFEC_DIFFR |= (1 << (channel_pt->channel_number));
		}
		if(channel_pt->dual_sample_hold){
			afec->AFEC_SHMR |= (1 << (channel_pt->channel_number));
		}
		
		adc_set_channel_gain(afec, channel_pt->channel_number, channel_pt->gain);
		
		adc_channel_enable(afec, channel_pt->channel_number);
		
		if(channel_pt->offset & AFEC_OFFSET_AUTO_0){
			channel_pt->offset = adc_tune_offset(afec, channel_pt->channel_number);
		}
		else{
			adc_channel_set_analog_offset(afec, channel_pt->channel_number, channel_pt->offset & 0x3FF);
		}
		
		
	}
	
	return 0;
}

void adc_set_channel_gain(Afec *const afec, adc_channel_num_t channel, adc_gain_t gain){
	afec->AFEC_CGR &= ~AFEC_CGR_msk(channel);
	afec->AFEC_CGR |= AFEC_CGR_set(channel, gain);
}


uint16_t adc_get_max_val(adc_resolution_t res, bool is_signed){
	uint32_t unsigned_max = 0;
	
	switch(res){
		case AFEC_12_BITS:
			unsigned_max = 4096;
		break;
		case AFEC_13_BITS:
			unsigned_max = 8192;
		break;
		case AFEC_14_BITS:
			unsigned_max = 16384;
		break;
		case AFEC_15_BITS:
			unsigned_max = 32768;
		break;
		case AFEC_16_BITS:
			unsigned_max = 65536;
		break;		
	}
	if(is_signed){
		return (uint16_t)(unsigned_max/2) - 1;
	}
	else{
		return (uint16_t)(unsigned_max - 1);
	}
}


void adc_change_sequence(Afec *const afec, bool enable_conversion_sequencing, adc_channel_num_t sequence[AFEC_NB_CHANNELS]){
	if(enable_conversion_sequencing){
		afec->AFEC_MR |= AFEC_MR_USEQ_REG_ORDER;
		afec->AFEC_SEQ1R = AFEC_SEQ1R_USCH0(sequence[0]);
		afec->AFEC_SEQ1R = AFEC_SEQ1R_USCH1(sequence[1]);
		afec->AFEC_SEQ1R = AFEC_SEQ1R_USCH2(sequence[2]);
		afec->AFEC_SEQ1R = AFEC_SEQ1R_USCH3(sequence[3]);
		afec->AFEC_SEQ1R = AFEC_SEQ1R_USCH4(sequence[4]);
		afec->AFEC_SEQ1R = AFEC_SEQ1R_USCH5(sequence[5]);
		afec->AFEC_SEQ1R = AFEC_SEQ1R_USCH6(sequence[6]);
		afec->AFEC_SEQ1R = AFEC_SEQ1R_USCH7(sequence[7]);
		afec->AFEC_SEQ2R = AFEC_SEQ2R_USCH8(sequence[8]);
		afec->AFEC_SEQ2R = AFEC_SEQ2R_USCH9(sequence[9]);
		afec->AFEC_SEQ2R = AFEC_SEQ2R_USCH10(sequence[10]);
		afec->AFEC_SEQ2R = AFEC_SEQ2R_USCH11(sequence[11]);
	}
	else{
		afec->AFEC_MR &= ~AFEC_MR_USEQ_REG_ORDER;
	}
}

uint16_t adc_tune_offset(Afec *afec, adc_channel_num_t channel){
	volatile int32_t res = 0xFFF;
	volatile int32_t lastRes = 0xFFF;
	 uint16_t offset = 0;
	do{
		lastRes = res;
		adc_channel_set_analog_offset(afec, channel, offset++);
		adc_start_conversion(afec);
		while(!adc_is_conversion_done(afec, channel));
		res = adc_get_result(afec, channel);
	}while(((abs(res) < abs(lastRes)) || abs(res) > 10) && offset < AFEC_MAX_OFFSET);
	
	return offset;
}

void adc_set_conversion_trigger(Afec *const afec, adc_trigger_t conversion_trigger){
	if(conversion_trigger != AFEC_TRIG_FREERUN){
		afec->AFEC_MR &= ~AFEC_MR_FREERUN_ON;
		afec->AFEC_MR |= AFEC_MR_TRGSEL(conversion_trigger);
	}
	else{
		afec->AFEC_MR &= ~(AFEC_MR_TRGSEL_Msk | AFEC_MR_TRGEN);
		afec->AFEC_MR |= AFEC_MR_FREERUN_ON;
	}
}

void (*adc_comp_callback_function[2])(void);

void adc_init_comparator(Afec *const afec, adc_comp_opts_t *const config, void (*callback)(void)){
	if(config->enable_comparator_interrupt){
		afec->AFEC_EMR |= AFEC_EMR_CMPMODE(config->compare_mode);
		afec->AFEC_EMR |= AFEC_EMR_CMPFILTER(config->event_filter);
		if(config->comp_channel == AFEC_CHANNEL_ALL){
			afec->AFEC_EMR |= AFEC_EMR_CMPALL;
		}
		else{
			afec->AFEC_EMR &= ~(AFEC_EMR_CMPALL | AFEC_EMR_CMPSEL_Msk);
			afec->AFEC_EMR |= AFEC_EMR_CMPSEL(config->comp_channel);
		}
		if(afec == AFEC0){
			adc_comp_callback_function[0] = callback;
			NVIC_EnableIRQ(AFEC0_IRQn);
		}
		else if(afec == AFEC1){
			adc_comp_callback_function[1] = callback;
			NVIC_EnableIRQ(AFEC1_IRQn);
		}
		
		afec->AFEC_IER |= AFEC_IER_COMPE;
	}
}

void adc_start_conversion(Afec *const afec){
	afec->AFEC_CR |= AFEC_CR_START;
}

bool adc_is_conversion_done(Afec *const afec, adc_channel_num_t channel){
	return !!(afec->AFEC_ISR & (1 << channel));
}

int16_t	adc_get_result(Afec *afec, adc_channel_num_t channel){	
	afec->AFEC_CSELR = channel;
	return afec->AFEC_CDR;
}

void AFEC0_Handler(){
	if(AFEC0->AFEC_ISR & (AFEC_ISR_COMPE) && adc_comp_callback_function[0]){
		adc_comp_callback_function[0]();
	}
}

void AFEC1_Handler(){
	if(AFEC1->AFEC_ISR & (AFEC_ISR_COMPE) && adc_comp_callback_function[1]){
		adc_comp_callback_function[1]();
	}
}

void adc_set_power_mode(Afec *const afec, const enum adc_power_mode mode)
{

	switch(mode) {
		case AFEC_POWER_MODE_ALL_ON:
			afec->AFEC_MR &= ~(AFEC_MR_FWUP_ON | AFEC_MR_SLEEP_SLEEP);
		break;
		case AFEC_POWER_MODE_REF_ON:
			afec->AFEC_MR |= (AFEC_MR_FWUP_ON | AFEC_MR_SLEEP_SLEEP);
		break;
		case AFEC_POWER_MODE_ALL_OFF:
			afec->AFEC_MR |= AFEC_MR_SLEEP_SLEEP;
			afec->AFEC_MR &= ~AFEC_MR_FWUP_ON;
		break;
	}
}

/**
 * \brief Enable AFEC Module.
 *
 * \param afec  Base address of the AFEC
 */
void adc_enable(Afec *const afec)
{
	if(afec == AFEC0){
		pmc_enable_periph_clk(AFEC0_IRQn);
	}
	if(afec == AFEC1){
		pmc_enable_periph_clk(AFEC1_IRQn);
	}
}

/**
 * \brief Disable AFEC Module.
 *
 * \param afec  Base address of the AFEC
 */
void adc_disable(Afec *const afec)
{
	if(afec == AFEC0){
		pmc_disable_periph_clk(AFEC0_IRQn);
	}
	if(afec == AFEC1){
		pmc_disable_periph_clk(AFEC1_IRQn);
	}
}

/**
 * \brief Get AFEC overrun error status.
 *
 * \param afec  Base address of the AFEC.
 *
 * \return AFEC overrun error status.
 */
static inline uint32_t adc_get_overrun_status(Afec *const afec)
{
	return afec->AFEC_OVER;
}

/**
 * \brief Enable the specified AFEC channel.
 *
 * \param afec  Base address of the AFEC.
 * \param adc_ch AFEC channel number.
 */
static void adc_channel_enable(Afec *const afec, adc_channel_num_t adc_ch)
{
	afec->AFEC_CHER = (adc_ch == AFEC_CHANNEL_ALL) ?
			AFEC_CHANNEL_ALL : 1 << adc_ch;
}

/**
 * \brief Disable the specified AFEC channel.
 *
 * \param afec  Base address of the AFEC.
 * \param adc_ch AFEC channel number.
 */
static inline void adc_channel_disable(Afec *const afec, const adc_channel_num_t adc_ch)
{

	afec->AFEC_CHDR = (adc_ch == AFEC_CHANNEL_ALL) ?
			AFEC_CHANNEL_ALL : 1 << adc_ch;
}

/**
 * \brief Get the AFEC channel status.
 *
 * \param afec  Base address of the AFEC.
 * \param adc_ch AFEC channel number.
 *
 * \retval 1 if channel is enabled.
 * \retval 0 if channel is disabled.
 */
static inline uint32_t adc_channel_get_status(Afec *const afec, const adc_channel_num_t adc_ch)
{
	return afec->AFEC_CHSR & (1 << adc_ch);
}

/**
 * \brief Set analog offset to be used for channel CSEL.
 *
 * \param afec  Base address of the AFEC.
 * \param adc_ch AFEC channel number.
 * \param aoffset  Analog offset value.
 */
static void adc_channel_set_analog_offset(Afec *const afec, adc_channel_num_t adc_ch, uint16_t aoffset)
{
	afec->AFEC_CSELR = adc_ch;
	afec->AFEC_COCR = AFEC_COCR_AOFF(aoffset);
}

/**
 * \brief Get the Last Data Converted.
 *
 * \param afec  Base address of the AFEC.
 *
 * \return AFEC latest converted value.
 */
static inline uint32_t adc_get_latest_value(Afec *const afec)
{
	return afec->AFEC_LCDR & AFEC_LCDR_LDATA_Msk;
}

/**
 * \brief Get the Last Converted Channel Number.
 *
 * \param afec  Base address of the AFEC.
 *
 * \return AFEC Last Converted Channel Number.
 */
static inline uint32_t adc_get_latest_chan_num(Afec *const afec)
{
	return (afec->AFEC_LCDR & AFEC_LCDR_CHNB_Msk) >> AFEC_LCDR_CHNB_Pos;
}

const volatile uint32_t *adc_get_result_register_address(Afec *afec){
	return &(afec->AFEC_LCDR);
}