/**
 * \file
 *
 * \brief Analog-Front-End Controller driver for SAM.
 *
 * Copyright (c) 2013-2015 Atmel Corporation. All rights reserved.
 *
 * \asf_license_start
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 *    Atmel microcontroller product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * \asf_license_stop
 *
 */
/*
 * Support and FAQ: visit <a href="http://www.atmel.com/design-support/">Atmel Support</a>
 */

#ifndef AFEC_H_INCLUDED
#define AFEC_H_INCLUDED


#include <stdbool.h>

/** Definitions for AFEC resolution */
typedef enum  {
	AFEC_12_BITS					= AFEC_EMR_RES_NO_AVERAGE,    /* AFEC 12-bit resolution */
	AFEC_13_BITS					= AFEC_EMR_RES_OSR4,          /* AFEC 13-bit resolution */
	AFEC_14_BITS					= AFEC_EMR_RES_OSR16,         /* AFEC 14-bit resolution */
	AFEC_15_BITS					= AFEC_EMR_RES_OSR64,         /* AFEC 15-bit resolution */
	AFEC_16_BITS					= AFEC_EMR_RES_OSR256         /* AFEC 16-bit resolution */
}adc_resolution_t;

/** Definitions for AFEC trigger */
typedef enum {
	AFEC_TRIG_SOFTWARE				= AFEC_MR_TRGEN_DIS,
	AFEC_TRIG_EXT					= AFEC_MR_TRGSEL_AFEC_TRIG0 | AFEC_MR_TRGEN,
	AFEC_TRIG_TIO_CH_0				= AFEC_MR_TRGSEL_AFEC_TRIG1 | AFEC_MR_TRGEN,
	AFEC_TRIG_TIO_CH_1				= AFEC_MR_TRGSEL_AFEC_TRIG2 | AFEC_MR_TRGEN,
	AFEC_TRIG_TIO_CH_2				= AFEC_MR_TRGSEL_AFEC_TRIG3 | AFEC_MR_TRGEN,
	AFEC_TRIG_PWM_EVENT_LINE_0		= AFEC_MR_TRGSEL_AFEC_TRIG4 | AFEC_MR_TRGEN,
	AFEC_TRIG_PWM_EVENT_LINE_1		= AFEC_MR_TRGSEL_AFEC_TRIG5 | AFEC_MR_TRGEN,
	AFEC_TRIG_ANALOG_COMPARATOR		= AFEC_MR_TRGSEL_AFEC_TRIG6 | AFEC_MR_TRGEN,
	AFEC_TRIG_FREERUN				= 0xFF,
}adc_trigger_t;

typedef enum{
	AFEC_SIGN_ALL_SIGNED			= AFEC_EMR_SIGNMODE_ALL_SIGNED,
	AFEC_SIGN_ALL_UNSIGNED			= AFEC_EMR_SIGNMODE_ALL_UNSIGNED,
	AFEC_SIGN_SINGLE_ENDED_SIGNED	= AFEC_EMR_SIGNMODE_SE_SIGN_DF_UNSG,
	AFEC_SIGN_SINGLE_ENDED_UNSIGNED = AFEC_EMR_SIGNMODE_SE_UNSG_DF_SIGN,
}adc_signMode_t;

/** Definitions for Comparison Mode */
typedef enum{
	AFEC_CMP_MODE_LOWER				= AFEC_EMR_CMPMODE_LOW,
	AFEC_CMP_MODE_HIGHER			= AFEC_EMR_CMPMODE_HIGH,
	AFEC_CMP_MODE_INSIDE			= AFEC_EMR_CMPMODE_IN,
	AFEC_CMP_MODE_OUTSIDE			= AFEC_EMR_CMPMODE_OUT
}adc_cmp_mode_t;

/** Definitions for AFEC channel number */
typedef enum  {
	AFEC_CHANNEL_0 = 0,
	AFEC_CHANNEL_1,
	AFEC_CHANNEL_2,
	AFEC_CHANNEL_3,
	AFEC_CHANNEL_4,
	AFEC_CHANNEL_5,
	AFEC_CHANNEL_6,
	AFEC_CHANNEL_7,
	AFEC_CHANNEL_8,
	AFEC_CHANNEL_9,
	AFEC_CHANNEL_10,
	AFEC_CHANNEL_11,
	AFEC_TEMPERATURE_SENSOR,
	AFEC_CHANNEL_ALL = 0x0FFF,
} adc_channel_num_t;

typedef enum{
	AFEC_GAIN_1 = 0,
	AFEC_GAIN_2 = 1,
	AFEC_GAIN_4 = 2
}adc_gain_t;

#define AFEC_NB_CHANNELS (12UL)
	
#define AFEC_OFFSET_AUTO_0 (1 << 15)

typedef struct{
	bool enable_comparator_interrupt;
	adc_cmp_mode_t compare_mode;
	uint16_t high_level;
	uint16_t low_level;
	adc_channel_num_t comp_channel;
	uint8_t event_filter;
}adc_comp_opts_t;

typedef struct{
	adc_channel_num_t channel_number;
	bool differential_channel;
	bool dual_sample_hold;
	uint16_t offset;
	adc_gain_t gain;
}adc_channel_opts_t;

typedef struct {
	adc_resolution_t resolution;
	adc_trigger_t conversion_trigger;
	adc_signMode_t sign_mode;
}adc_opts_t;



uint32_t	adc_init(Afec *const afec, adc_opts_t *config, adc_channel_opts_t enabled_channels[], uint8_t num_channels);
void		adc_change_sequence(Afec *const afec, bool enable_conversion_sequencing, adc_channel_num_t sequence[AFEC_NB_CHANNELS]);
void		adc_init_comparator(Afec *const afec, adc_comp_opts_t *const config, void (*callback)(void));

void		adc_start_conversion(Afec *const afec);
bool		adc_is_conversion_done(Afec *const afec, adc_channel_num_t channel);
int16_t		adc_get_result(Afec *afec, adc_channel_num_t channel);

const volatile uint32_t *adc_get_result_register_address(Afec *afec);


#endif /* AFEC_H_INCLUDED */
