
#include <sam.h>
#include "can_types.h"
#include "mcan.h"

#include "../pmc.h"
#include "../pio.h"

#include <stdlib.h>

void (*mcan_callback_function[2])(uint32_t index);

static void mcan_setBaudrate(Mcan *mcan, uint32_t baudrate);
static void mcan_switch_to_upll(uint32_t prescaler);
static void mcan_rx_fifo_acknowledge(Mcan *mcan, bool fifo_number, uint32_t index);
static void mcan_clear_interrupt_status(Mcan *mcan, const enum mcan_interrupt_source source);
static void mcan_setConfiguration(Mcan *mcan, struct mcan_config *config);
static void mcan_messageMemoryInit(Mcan *mcan);
static void mcan_enablePeripheralClock(Mcan *mcan);
static uint32_t mcan_read_interrupt_status(Mcan *mcan);

// Index counter
volatile uint32_t mcan0_fifo0_index = 0;
volatile uint32_t mcan1_fifo0_index = 0;

//********************************************
//************* DEFINES **********************
//********************************************
/* PCK5 ID,assigned to MCAN module */

#define TSEG1		10
#define TSEG2		7

#define BRP_1000K	3
#define SJW_1000K	3

#define BRP_500K	7
#define SJW_500K	3

#define BRP_250K	15
#define SJW_250K	3

#define COMPILER_ALIGNED(a)    __attribute__((__aligned__(a)))
// MCAN 0
COMPILER_ALIGNED(4)
static struct mcan_rx_element_fifo_0 mcan0_rx_fifo_0[MCAN0_RX_FIFO_0_NUM];
COMPILER_ALIGNED(4)
static struct mcan_tx_element mcan0_tx_buffer[MCAN0_TX_FIFO_QUEUE_NUM];
COMPILER_ALIGNED(4)
static struct mcan_standard_message_filter_element mcan0_rx_standard_filter[MCAN0_RX_STANDARD_ID_FILTER_NUM];
COMPILER_ALIGNED(4)
static struct mcan_extended_message_filter_element mcan0_rx_extended_filter[MCAN0_RX_EXTENDED_ID_FILTER_NUM];
// MCAN 1
COMPILER_ALIGNED(4)
static struct mcan_rx_element_fifo_0 mcan1_rx_fifo_0[MCAN1_RX_FIFO_0_NUM];
COMPILER_ALIGNED(4)
static struct mcan_tx_element mcan1_tx_buffer[MCAN1_TX_FIFO_QUEUE_NUM];
COMPILER_ALIGNED(4)
static struct mcan_standard_message_filter_element mcan1_rx_standard_filter[MCAN1_RX_STANDARD_ID_FILTER_NUM];
COMPILER_ALIGNED(4)
static struct mcan_extended_message_filter_element mcan1_rx_extended_filter[MCAN1_RX_EXTENDED_ID_FILTER_NUM];
//***********************************************************************************************************

void mcan_init(Mcan *mcan, uint32_t baudRate) {
	//MCAN0
	if (mcan == MCAN0) {
		pio_setMux(PIOB,3,A);
		pio_setMux(PIOB,2,A);
	}
	else if (mcan == MCAN1) {
		//MCAN1
		#ifdef __SAMV71Q21__
			pio_setMux(PIOC,14,C);
			pio_setMux(PIOC,12,C);
		#endif
		#ifdef __SAME70N21__
			pio_setMux(PIOD,28,B);
			pio_setMux(PIOD,12,B);
		#endif
	}
	// UPLL has to be set up to 480 MHz. Div by 6 to achieve 80 MHz Can clock
	mcan_switch_to_upll(6);
	mcan_enablePeripheralClock(mcan);
	
	/* Configuration Change Enable. */
	mcan->MCAN_CCCR |= MCAN_CCCR_CCE;
	/* Initialize the message memory address. */
	mcan_messageMemoryInit(mcan);
	
	 struct mcan_config config;
	mcan_get_config_defaults(&config);
	mcan_setConfiguration(mcan, &config);
	mcan_setBaudrate(mcan, baudRate);
	
	/* Enable the interrupt settings. */
	mcan->MCAN_ILE = MCAN_ILE_EINT0 | MCAN_ILE_EINT1;
	mcan->MCAN_TXBTIE = 0xFFFFFFFFul;
	mcan->MCAN_TXBCIE = 0xFFFFFFFFul;
	
	mcan->MCAN_CCCR &= ~MCAN_CCCR_INIT;
	/* Wait for the sync. */
	while (mcan->MCAN_CCCR & MCAN_CCCR_INIT);
	// Now CAN has started
	
}

void mcan_setExtendedFilter(Mcan *mcan, uint32_t *extendedIdList, uint32_t nr_Ids) {
	uint32_t i;
	if (mcan == MCAN0) {
		if (nr_Ids == 0) {
			mcan0_rx_extended_filter[0].F0.bit.EFEC		= 1; // Store in Rx FIFO 0 if filter matches
			mcan0_rx_extended_filter[0].F0.bit.EFID1	= 0; // First ID of range filter
			mcan0_rx_extended_filter[0].F1.bit.EFT		= 0; // Range Filter
			mcan0_rx_extended_filter[0].F1.bit.EFID2	= 0x1FFFFFFF; //  536870911; // End ID range filter
		}
		else {
			// Odd amount of Ids
			if (nr_Ids % 2) {
				for (i = 0; i < (nr_Ids-1)/2; i++) {
					mcan0_rx_extended_filter[i].F0.bit.EFEC = 1; // Store in Rx FIFO 0 if filter matches
					mcan0_rx_extended_filter[i].F1.bit.EFT	= 1; // Dual ID filter
					mcan0_rx_extended_filter[i].F0.bit.EFID1 = extendedIdList[i*2];
					mcan0_rx_extended_filter[i].F1.bit.EFID2 = extendedIdList[i*2+1];
				}
				mcan0_rx_extended_filter[i].F0.bit.EFEC = 1; // Store in Rx FIFO 0 if filter matches
				mcan0_rx_extended_filter[i].F1.bit.EFT	= 1; // Dual ID filter
				mcan0_rx_extended_filter[i].F0.bit.EFID1 = extendedIdList[nr_Ids-1];
				mcan0_rx_extended_filter[i].F1.bit.EFID2 = extendedIdList[nr_Ids-1];
			}
			// Even amount of Ids
			else {
				for (i = 0; i < nr_Ids/2; i++) {
					mcan0_rx_extended_filter[i].F0.bit.EFEC = 1; // Store in Rx FIFO 0 if filter matches
					mcan0_rx_extended_filter[i].F1.bit.EFT	= 1; // Dual ID filter
					mcan0_rx_extended_filter[i].F0.bit.EFID1 = extendedIdList[i*2];
					mcan0_rx_extended_filter[i].F1.bit.EFID2 = extendedIdList[i*2+1];
				}
			}
		}
	}
	else if (mcan == MCAN1) {
		if (nr_Ids == 0) {
			mcan1_rx_extended_filter[0].F0.bit.EFEC		= 1; // Store in Rx FIFO 0 if filter matches
			mcan1_rx_extended_filter[0].F0.bit.EFID1	= 0; // First ID of range filter
			mcan1_rx_extended_filter[0].F1.bit.EFT		= 0; // Range Filter
			mcan1_rx_extended_filter[0].F1.bit.EFID2	= 0x1FFFFFFF; // End ID range filter
		}
		else {
			// Odd amount of Ids
			if (nr_Ids % 2) {
				for (i = 0; i < (nr_Ids-1)/2; i++) {
					mcan1_rx_extended_filter[i].F0.bit.EFEC = 1; // Store in Rx FIFO 0 if filter matches
					mcan1_rx_extended_filter[i].F1.bit.EFT	= 1; // Dual ID filter
					mcan1_rx_extended_filter[i].F0.bit.EFID1 = extendedIdList[i*2];
					mcan1_rx_extended_filter[i].F1.bit.EFID2 = extendedIdList[i*2+1];
				}
				mcan1_rx_extended_filter[i].F0.bit.EFEC = 1; // Store in Rx FIFO 0 if filter matches
				mcan1_rx_extended_filter[i].F1.bit.EFT	= 1; // Dual ID filter
				mcan1_rx_extended_filter[i].F0.bit.EFID1 = extendedIdList[nr_Ids-1];
				mcan1_rx_extended_filter[i].F1.bit.EFID2 = extendedIdList[nr_Ids-1];
			}
			// Even amount of Ids
			else {
				for (i = 0; i < nr_Ids/2; i++) {
					mcan1_rx_extended_filter[i].F0.bit.EFEC = 1; // Store in Rx FIFO 0 if filter matches
					mcan1_rx_extended_filter[i].F1.bit.EFT	= 1; // Dual ID filter
					mcan1_rx_extended_filter[i].F0.bit.EFID1 = extendedIdList[i*2];
					mcan1_rx_extended_filter[i].F1.bit.EFID2 = extendedIdList[i*2+1];
				}
			}
		}
	}
}

void mcan_setStandardFilter(Mcan *mcan, uint32_t *standardIdList, uint32_t nr_Ids) {
	uint32_t i;
	if (mcan == MCAN0) {
		if (nr_Ids == 0) {
			mcan0_rx_standard_filter[0].S0.bit.SFT = 0; // Range filter (SFID2 to SFID1)
			mcan0_rx_standard_filter[0].S0.bit.SFEC = 1; // Store in RX FIFO 0
			mcan0_rx_standard_filter[0].S0.bit.SFID1 = 1;
			mcan0_rx_standard_filter[0].S0.bit.SFID2 = 2047;
		}
		else {
			// Odd amount of Ids
			if (nr_Ids % 2) {
				for (i = 0; i < (nr_Ids-1)/2; i++) {
					mcan0_rx_standard_filter[i].S0.bit.SFT = 1; // Dual ID filter
					mcan0_rx_standard_filter[i].S0.bit.SFEC = 1; // Store in RX FIFO 0
					mcan0_rx_standard_filter[i].S0.bit.SFID1 = standardIdList[i*2];
					mcan0_rx_standard_filter[i].S0.bit.SFID2 = standardIdList[i*2+1];
				}
				mcan0_rx_standard_filter[(nr_Ids-1)/2].S0.bit.SFT = 1; // Dual ID filter
				mcan0_rx_standard_filter[(nr_Ids-1)/2].S0.bit.SFEC = 1; // Store in RX FIFO 0
				mcan0_rx_standard_filter[(nr_Ids-1)/2].S0.bit.SFID1 = standardIdList[nr_Ids-1];
				mcan0_rx_standard_filter[(nr_Ids-1)/2].S0.bit.SFID2 = standardIdList[nr_Ids-1];
				
			}
			// Even amount of Ids
			else {
				for (i = 0; i < nr_Ids/2; i++) {
					mcan0_rx_standard_filter[i].S0.bit.SFT = 1; // Dual ID filter
					mcan0_rx_standard_filter[i].S0.bit.SFEC = 1; // Store in RX FIFO 0
					mcan0_rx_standard_filter[i].S0.bit.SFID1 = standardIdList[i*2];
					mcan0_rx_standard_filter[i].S0.bit.SFID2 = standardIdList[i*2+1];
				}
			}
		}
	}
	else if (mcan == MCAN1) {
		if (nr_Ids == 0) {
			mcan1_rx_standard_filter[0].S0.bit.SFT = 0; // Range filter (SFID2 to SFID1)
			mcan1_rx_standard_filter[0].S0.bit.SFEC = 1; // Store in RX FIFO 0
			mcan1_rx_standard_filter[0].S0.bit.SFID1 = 1;
			mcan1_rx_standard_filter[0].S0.bit.SFID2 = 2047;
		}
		else {
			// Odd amount of Ids
			if (nr_Ids % 2) {
				for (i = 0; i < (nr_Ids-1)/2; i++) {
					mcan1_rx_standard_filter[i].S0.bit.SFT = 1; // Dual ID filter
					mcan1_rx_standard_filter[i].S0.bit.SFEC = 1; // Store in RX FIFO 0
					mcan1_rx_standard_filter[i].S0.bit.SFID1 = standardIdList[i*2];
					mcan1_rx_standard_filter[i].S0.bit.SFID2 = standardIdList[i*2+1];
				}
				mcan1_rx_standard_filter[(nr_Ids-1)/2].S0.bit.SFT = 1; // Dual ID filter
				mcan1_rx_standard_filter[(nr_Ids-1)/2].S0.bit.SFEC = 1; // Store in RX FIFO 0
				mcan1_rx_standard_filter[(nr_Ids-1)/2].S0.bit.SFID1 = standardIdList[nr_Ids-1];
				mcan1_rx_standard_filter[(nr_Ids-1)/2].S0.bit.SFID2 = standardIdList[nr_Ids-1];
				
			}
			// Even amount of Ids
			else {
				for (i = 0; i < nr_Ids/2; i++) {
					mcan1_rx_standard_filter[i].S0.bit.SFT = 1; // Dual ID filter
					mcan1_rx_standard_filter[i].S0.bit.SFEC = 1; // Store in RX FIFO 0
					mcan1_rx_standard_filter[i].S0.bit.SFID1 = standardIdList[i*2];
					mcan1_rx_standard_filter[i].S0.bit.SFID2 = standardIdList[i*2+1];
				}
			}
		}
	}
}



enum status_code mcan_sendMessage(Mcan *mcan, struct Can_message_t *message) {
	uint32_t index = 0;
	uint32_t i;
	if (message->idType == EXTENDED_CAN_ID) {
		// If FIFO full
		if (mcan->MCAN_TXFQS & MCAN_TXFQS_TFQF) {
			return ERR_BUSY;
		}
		index = (mcan->MCAN_TXFQS & MCAN_TXFQS_TFQPI_Msk) >> MCAN_TXFQS_TFQPI_Pos;
		if (mcan == MCAN0) {
			
			mcan0_tx_buffer[index].T0.bit.RTR = 0;
			mcan0_tx_buffer[index].T0.bit.XTD = 1;
			mcan0_tx_buffer[index].T1.bit.EFCC = 0;
			mcan0_tx_buffer[index].T1.bit.MM = 0;
			
			//mcan0_tx_buffer[index].T0.bit.ID = message->messageID << MCAN_TX_ELEMENT_T0_STANDARD_ID_Pos;
			mcan0_tx_buffer[index].T0.bit.ID = message->messageID;
			mcan0_tx_buffer[index].T1.bit.DLC = message->dataLength;
			for (i = 0; i < MCAN_ELEMENT_DATA_SIZE; i++) {
				mcan0_tx_buffer[index].data[i] = message->data.u8[i];
			}
		}
		else if (mcan == MCAN1) {
			mcan1_tx_buffer[index].T0.bit.RTR = 0;
			mcan1_tx_buffer[index].T0.bit.XTD = 1;
			mcan1_tx_buffer[index].T1.bit.EFCC = 0;
			mcan1_tx_buffer[index].T1.bit.MM = 0;
			
			mcan1_tx_buffer[index].T0.bit.ID = message->messageID;
			mcan1_tx_buffer[index].T1.bit.DLC = message->dataLength;
			for (i = 0; i < MCAN_ELEMENT_DATA_SIZE; i++) {
				mcan1_tx_buffer[index].data[i] = message->data.u8[i];
			}
		}
		if (mcan->MCAN_CCCR & MCAN_CCCR_CCE) {
			return ERR_BUSY;
		}
		mcan->MCAN_TXBAR = (1<<index);
		return STATUS_OK;
	}
	else if (message->idType == STANDARD_CAN_ID) {
		// index  CAN messages memory index for transmitting CAN ID

		// If FIFO full
		if (mcan->MCAN_TXFQS & MCAN_TXFQS_TFQF) {
			return ERR_BUSY;
		}
		index = (mcan->MCAN_TXFQS & MCAN_TXFQS_TFQPI_Msk) >> MCAN_TXFQS_TFQPI_Pos;
		if (mcan == MCAN0) {
	
			mcan0_tx_buffer[index].T0.bit.RTR = 0;
			mcan0_tx_buffer[index].T0.bit.XTD = 0;
			mcan0_tx_buffer[index].T1.bit.EFCC = 0;
			mcan0_tx_buffer[index].T1.bit.MM = 0;
	
			mcan0_tx_buffer[index].T0.bit.ID = message->messageID << MCAN_TX_ELEMENT_T0_STANDARD_ID_Pos;
			mcan0_tx_buffer[index].T1.bit.DLC = message->dataLength;
			for (i = 0; i < MCAN_ELEMENT_DATA_SIZE; i++) {
				mcan0_tx_buffer[index].data[i] = message->data.u8[i];
			}
		}
		else if (mcan == MCAN1) {
			mcan1_tx_buffer[index].T0.bit.RTR = 0;
			mcan1_tx_buffer[index].T0.bit.XTD = 0;
			mcan1_tx_buffer[index].T1.bit.EFCC = 0;
			mcan1_tx_buffer[index].T1.bit.MM = 0;
	
			mcan1_tx_buffer[index].T0.bit.ID = message->messageID << MCAN_TX_ELEMENT_T0_STANDARD_ID_Pos;
			mcan1_tx_buffer[index].T1.bit.DLC = message->dataLength;
			for (i = 0; i < MCAN_ELEMENT_DATA_SIZE; i++) {
				mcan1_tx_buffer[index].data[i] = message->data.u8[i];
			}
		}
		if (mcan->MCAN_CCCR & MCAN_CCCR_CCE) {
			return ERR_BUSY;
		}
		mcan->MCAN_TXBAR = (1<<index);
		return STATUS_OK;		
	}
	return ERR_BAD_DATA;
}

void mcan_getMessage(Mcan *mcan, struct Can_message_t *message, uint32_t index) {
	uint32_t i;
	if (mcan == MCAN0) {
		// Check for extended ID
		if (mcan0_rx_fifo_0[index].R0.bit.XTD == 1) {
			message->messageID	= mcan0_rx_fifo_0[index].R0.bit.ID;
			message->dataLength = mcan0_rx_fifo_0[index].R1.bit.DLC;
			for (i = 0 ; i < message->dataLength ; i++){
				message->data.u8[i]	= mcan0_rx_fifo_0[index].data[i];
			}
		}
		else {
			message->messageID	= mcan0_rx_fifo_0[index].R0.bit.ID >> MCAN_TX_ELEMENT_T0_STANDARD_ID_Pos;
			message->dataLength = mcan0_rx_fifo_0[index].R1.bit.DLC;
			for (i = 0 ; i < message->dataLength ; i++){
				message->data.u8[i]	= mcan0_rx_fifo_0[index].data[i];
			}	
		}
	}
	else if (mcan == MCAN1) {
		// Check for extended ID
		if (mcan1_rx_fifo_0[index].R0.bit.XTD == 1) {
			message->messageID	= mcan1_rx_fifo_0[index].R0.bit.ID;
			message->dataLength = mcan1_rx_fifo_0[index].R1.bit.DLC;
			for (i = 0 ; i < message->dataLength ; i++){
				message->data.u8[i]	= mcan1_rx_fifo_0[index].data[i];
			}
		}
		else {
			message->messageID	= mcan1_rx_fifo_0[index].R0.bit.ID >> MCAN_TX_ELEMENT_T0_STANDARD_ID_Pos;
			message->dataLength = mcan1_rx_fifo_0[index].R1.bit.DLC;
			for (i = 0 ; i < message->dataLength ; i++){
				message->data.u8[i]	= mcan1_rx_fifo_0[index].data[i];
			}
		}
	}
}

void MCAN0_Handler(){
	volatile uint32_t status;
	status = mcan_read_interrupt_status(MCAN0);
	if (status & MCAN_RX_FIFO_0_FULL) {
		mcan_clear_interrupt_status(MCAN0, MCAN_RX_FIFO_0_FULL);
		mcan0_fifo0_index = mcan0_fifo0_index + 1;
		if (mcan0_fifo0_index >= MCAN0_RX_FIFO_0_NUM) {
			mcan0_fifo0_index = 0;
		}
	}	
	if (status & MCAN_RX_FIFO_0_NEW_MESSAGE) {
		mcan_clear_interrupt_status(MCAN0, MCAN_RX_FIFO_0_NEW_MESSAGE);
		if (mcan_callback_function[0] != NULL) {
			(*mcan_callback_function[0])(mcan0_fifo0_index);
		}
		mcan_rx_fifo_acknowledge(MCAN0, 0, mcan0_fifo0_index);
		mcan0_fifo0_index = mcan0_fifo0_index + 1;
		if (mcan0_fifo0_index == MCAN0_RX_FIFO_0_NUM) {
			mcan0_fifo0_index = 0;
		}
	}
}

void MCAN1_Handler(){
	volatile uint32_t status;
	status = mcan_read_interrupt_status(MCAN1);
	if (status & MCAN_RX_FIFO_0_FULL) {
		mcan_clear_interrupt_status(MCAN1, MCAN_RX_FIFO_0_FULL);
		mcan1_fifo0_index = mcan1_fifo0_index + 1;
		if (mcan1_fifo0_index >= MCAN1_RX_FIFO_0_NUM) {
			mcan1_fifo0_index = 0;
		}
	}
	if (status & MCAN_RX_FIFO_0_NEW_MESSAGE) {
		mcan_clear_interrupt_status(MCAN1, MCAN_RX_FIFO_0_NEW_MESSAGE);
		if (mcan_callback_function[1] != NULL) {
			(*mcan_callback_function[1])(mcan1_fifo0_index);
		}
		mcan_rx_fifo_acknowledge(MCAN1, 0, mcan1_fifo0_index);
		mcan1_fifo0_index = mcan1_fifo0_index + 1;
		if (mcan1_fifo0_index == MCAN1_RX_FIFO_0_NUM) {
			mcan1_fifo0_index = 0;
		}
	}
}

void mcan_rx_fifo_acknowledge(Mcan *mcan, bool fifo_number, uint32_t index)
{
	if (!fifo_number) {
		mcan->MCAN_RXF0A = MCAN_RXF0A_F0AI(index);
	} 
	else {
		mcan->MCAN_RXF1A = MCAN_RXF1A_F1AI(index);
	}
}



static void mcan_enablePeripheralClock(Mcan *mcan)
{
	if (mcan == MCAN0) {
		/* Turn on the digital interface clock. */
		pmc_enable_periph_clk(ID_MCAN0);
	} 
	else if (mcan == MCAN1) {
		/* Turn on the digital interface clock. */
		pmc_enable_periph_clk(ID_MCAN1);
	}
}

void mcan_switch_to_upll(uint32_t prescaler){
	// Disables clock to MCAN
	pmc_disable_programmable_clk(PMC_PCK_5);
	// Register sets pres + 1
	pmc_switch_programmable_clk_to_upllck(PMC_PCK_5, PMC_PCK_PRES(prescaler-1));
	pmc_enable_programmable_clk(PMC_PCK_5);
}

static void mcan_setBaudrate(Mcan *mcan, uint32_t baudrate)
{
	// Mcan clock is 80 MHz
	// Time Quanta Frequency = canClock (80MHz) / (BRP + 1)
	// Number of time Quantas = (TSEG1 + TSEG2 + 3)
	// Bit rate is = Time Quanta Frequency / Number of time Quantas
	
	switch (baudrate) {
		case CAN_BPS_1000K:
			mcan->MCAN_BTP = (MCAN_BTP_BRP(BRP_1000K) | MCAN_BTP_SJW(SJW_1000K) | MCAN_BTP_TSEG1(TSEG1) | MCAN_BTP_TSEG2(TSEG2) );
		break;
		case CAN_BPS_500K:
			mcan->MCAN_BTP = (MCAN_BTP_BRP(BRP_500K) | MCAN_BTP_SJW(SJW_500K) | MCAN_BTP_TSEG1(TSEG1) | MCAN_BTP_TSEG2(TSEG2) );
		break;
		case CAN_BPS_250K:
			mcan->MCAN_BTP = (MCAN_BTP_BRP(BRP_250K) | MCAN_BTP_SJW(SJW_250K) | MCAN_BTP_TSEG1(TSEG1) | MCAN_BTP_TSEG2(TSEG2) );
		break;
	}
}


void mcan_enable_interrupt(Mcan *mcan, void (*callback_function)(uint32_t index), uint32_t interrupt_priority)
{
	if (callback_function == NULL) {
		return;
	}
	else {
		if (mcan == MCAN0) {
			mcan_callback_function[0] = callback_function;
			NVIC_DisableIRQ(MCAN0_IRQn);
			NVIC_ClearPendingIRQ(MCAN0_IRQn);
			NVIC_SetPriority(MCAN0_IRQn,interrupt_priority); // must be >= 5
			NVIC_EnableIRQ(MCAN0_IRQn);
		}
		else if (mcan == MCAN1){
			mcan_callback_function[1] = callback_function;
			NVIC_DisableIRQ(MCAN1_IRQn);
			NVIC_ClearPendingIRQ(MCAN1_IRQn);
			NVIC_SetPriority(MCAN1_IRQn,interrupt_priority); // must be >= 5
			NVIC_EnableIRQ(MCAN1_IRQn);
		}
		mcan->MCAN_IE |= MCAN_RX_FIFO_0_NEW_MESSAGE | MCAN_RX_FIFO_0_FULL;
	}
}

void mcan_disable_interrupt(Mcan *mcan)
{
	if (mcan == MCAN0) {
		NVIC_DisableIRQ(MCAN0_IRQn);
	}
	else if (mcan == MCAN1) {
		NVIC_DisableIRQ(MCAN1_IRQn);
	}
	mcan->MCAN_IE &= ~MCAN_RX_FIFO_0_NEW_MESSAGE;
}

static uint32_t mcan_read_interrupt_status(Mcan *mcan) {
	return mcan->MCAN_IR;
}

static void mcan_clear_interrupt_status(Mcan *mcan, const enum mcan_interrupt_source source)
{
	mcan->MCAN_IR = source;
}

static void mcan_messageMemoryInit(Mcan *mcan)
{
	if (mcan == MCAN0) {
		mcan->MCAN_SIDFC |= (((uint32_t)mcan0_rx_standard_filter) & 0x0000fffc) |
		MCAN_SIDFC_LSS(MCAN0_RX_STANDARD_ID_FILTER_NUM);
		
		mcan->MCAN_RXF0C |= (((uint32_t)mcan0_rx_fifo_0) & 0x0000fffc) |
		MCAN_RXF0C_F0S(MCAN0_RX_FIFO_0_NUM);
		
		mcan->MCAN_TXBC |= (((uint32_t)mcan0_tx_buffer) & 0x0000fffc) |
		MCAN_TXBC_TFQS(MCAN0_TX_FIFO_QUEUE_NUM);
		
		//EXTENDED CONFIG
		mcan->MCAN_XIDFC |= (((uint32_t)mcan0_rx_extended_filter) & 0x0000fffc) |
		MCAN_XIDFC_LSE(MCAN0_RX_EXTENDED_ID_FILTER_NUM);
		
		
	}
	else if (mcan == MCAN1) {
		mcan->MCAN_SIDFC |= (((uint32_t)mcan1_rx_standard_filter) & 0x0000fffc) |
		MCAN_SIDFC_LSS(MCAN1_RX_STANDARD_ID_FILTER_NUM);
		
		mcan->MCAN_RXF0C |= (((uint32_t)mcan1_rx_fifo_0) & 0x0000fffc) |
		MCAN_RXF0C_F0S(MCAN1_RX_FIFO_0_NUM);
		
		mcan->MCAN_TXBC |= (((uint32_t)mcan1_tx_buffer) & 0x0000fffc)|
		MCAN_TXBC_TFQS(MCAN1_TX_FIFO_QUEUE_NUM);
		
		//EXTENDED CONFIG
		mcan->MCAN_XIDFC |= (((uint32_t)mcan1_rx_extended_filter) & 0x0000fffc) |
		MCAN_XIDFC_LSE(MCAN1_RX_EXTENDED_ID_FILTER_NUM);
	}
	mcan->MCAN_RXESC = 0; // 8 byte data field
}

static void mcan_setConfiguration(Mcan *mcan, struct mcan_config *config)
{
	/* Timing setting. */

	mcan->MCAN_RWD |= MCAN_RWD_WDC(config->watchdog_configuration);

	if (config->transmit_pause) {
		mcan->MCAN_CCCR |= MCAN_CCCR_TXP;
	}

	if (!config->automatic_retransmission) {
		mcan->MCAN_CCCR |= MCAN_CCCR_DAR;
	}

	if (config->clock_stop_request) {
		mcan->MCAN_CCCR |= MCAN_CCCR_CSR;
	}
	mcan->MCAN_TSCC = MCAN_TSCC_TCP(config->timestamp_prescaler) |
	MCAN_TSCC_TSS_TCP_INC;

	mcan->MCAN_TOCC = MCAN_TOCC_TOP(config->timeout_period) |
	config->timeout_mode | config->timeout_enable;

	mcan->MCAN_GFC = MCAN_GFC_ANFS(config->nonmatching_frames_action_standard) |
	MCAN_GFC_ANFE(config->nonmatching_frames_action_extended);
	if (config->remote_frames_standard_reject) {
		mcan->MCAN_GFC |= MCAN_GFC_RRFS;
	}
	if (config->remote_frames_extended_reject) {
		mcan->MCAN_GFC|= MCAN_GFC_RRFE;
	}

	mcan->MCAN_XIDAM = config->extended_id_mask;

	if (config->rx_fifo_0_overwrite) {
		mcan->MCAN_RXF0C |= MCAN_RXF0C_F0OM;
	}
	mcan->MCAN_RXF0C |= MCAN_RXF0C_F0WM(config->rx_fifo_0_watermark);

	if (config->rx_fifo_1_overwrite) {
		mcan->MCAN_RXF1C |= MCAN_RXF1C_F1OM;
	}
	mcan->MCAN_RXF1C |= MCAN_RXF1C_F1WM(config->rx_fifo_1_watermark);

	if (config->tx_queue_mode) {
		mcan->MCAN_TXBC |= MCAN_TXBC_TFQM;
	}

	mcan->MCAN_TXEFC |= MCAN_TXEFC_EFWM(config->tx_event_fifo_watermark);
}




// void mcan_stop(Mcan *mcan)
// {
// 	mcan->MCAN_CCCR |= MCAN_CCCR_INIT;
// 	/* Wait for the sync. */
// 	while (!(mcan->MCAN_CCCR & MCAN_CCCR_INIT));
// }
/*
enum status_code mcan_sendExtendedMessage(Mcan *mcan, struct Can_message_t message) {
	// index  CAN messages memory index for transmitting CAN ID
	uint32_t index = 0;
	uint32_t i;
	// If FIFO full
	if (mcan->MCAN_TXFQS & MCAN_TXFQS_TFQF) {
		return ERR_BUSY;
	}
	index = (mcan->MCAN_TXFQS & MCAN_TXFQS_TFQPI_Msk) >> MCAN_TXFQS_TFQPI_Pos;
	if (mcan == MCAN0) {
		
		mcan0_tx_buffer[index].T0.bit.RTR = 0;
		mcan0_tx_buffer[index].T0.bit.XTD = 1;
		mcan0_tx_buffer[index].T1.bit.EFCC = 0;
		mcan0_tx_buffer[index].T1.bit.MM = 0;
		
		//mcan0_tx_buffer[index].T0.bit.ID = message.messageID << MCAN_TX_ELEMENT_T0_STANDARD_ID_Pos;
		mcan0_tx_buffer[index].T0.bit.ID = message.messageID;
		mcan0_tx_buffer[index].T1.bit.DLC = message.dataLength;
		for (i = 0; i < MCAN_ELEMENT_DATA_SIZE; i++) {
			mcan0_tx_buffer[index].data[i] = message.data.u8[i];
		}
	}
	else if (mcan == MCAN1) {
		mcan1_tx_buffer[index].T0.bit.RTR = 0;
		mcan1_tx_buffer[index].T0.bit.XTD = 1;
		mcan1_tx_buffer[index].T1.bit.EFCC = 0;
		mcan1_tx_buffer[index].T1.bit.MM = 0;
		
		mcan1_tx_buffer[index].T0.bit.ID = message.messageID;
		mcan1_tx_buffer[index].T1.bit.DLC = message.dataLength;
		for (i = 0; i < MCAN_ELEMENT_DATA_SIZE; i++) {
			mcan1_tx_buffer[index].data[i] = message.data.u8[i];
		}
	}
	if (mcan->MCAN_CCCR & MCAN_CCCR_CCE) {
		return ERR_BUSY;
	}
	mcan->MCAN_TXBAR = (1<<index);
	return STATUS_OK;
	
}
*/