
#include <stddef.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <sam.h>

#include "interrupt_util.h"
#include "eefc.h"

void system_init_flash( uint32_t ul_clk )
{
  /* Set FWS for embedded Flash access according to operating frequency */  
  if ( ul_clk < CHIP_FREQ_FWS_0 )
  {
    EFC->EEFC_FMR = EEFC_FMR_FWS(0)|EEFC_FMR_CLOE;
  }
  else
  {
    if (ul_clk < CHIP_FREQ_FWS_1)
    {
      EFC->EEFC_FMR = EEFC_FMR_FWS(1)|EEFC_FMR_CLOE;
    }
    else
    {
      if (ul_clk < CHIP_FREQ_FWS_2)
      {
        EFC->EEFC_FMR = EEFC_FMR_FWS(2)|EEFC_FMR_CLOE;
      }
      else
      {
        if ( ul_clk < CHIP_FREQ_FWS_3 )
        {
          EFC->EEFC_FMR = EEFC_FMR_FWS(3)|EEFC_FMR_CLOE;
        }
        else
        {
          if ( ul_clk < CHIP_FREQ_FWS_4 )
          {
            EFC->EEFC_FMR = EEFC_FMR_FWS(4)|EEFC_FMR_CLOE;
          }
          else
          {
            EFC->EEFC_FMR = EEFC_FMR_FWS(5)|EEFC_FMR_CLOE;
          }
        }
      }
    }
  }
}
    
_NO_INLINE _RAMFUNC uint32_t eefc_perform_read_sequence(Efc *p_efc,
		uint32_t cmd_start, uint32_t cmd_stop,
		uint32_t *buf_out, size_t size)
{
	volatile uint32_t ul_status;
	uint32_t ul_cnt;

	uint32_t *p_ul_data = (uint32_t *) IFLASH_ADDR;
	
	if (buf_out == NULL) {
		return EFC_RC_INVALID;
	}
	
	do {
		ul_status = p_efc->EEFC_FSR;
	} while (!(ul_status & EEFC_FSR_FRDY));
	
	p_efc->EEFC_FMR |= (0x1u << 16);

	/* Send the Start Read command */
	p_efc->EEFC_FCR = EEFC_FCR_FKEY_PASSWD | EEFC_FCR_FARG(0)
			| EEFC_FCR_FCMD(cmd_start);

	/* Wait for the FRDY bit in the Flash Programming Status Register
	 * (EEFC_FSR) falls.
	 */
	do {
		ul_status = p_efc->EEFC_FSR;
	} while ((ul_status & EEFC_FSR_FRDY));

	/* The data is located in the first address of the Flash
	 * memory mapping.
	 */
	for (ul_cnt = 0; ul_cnt < size; ul_cnt++) {
		buf_out[ul_cnt] = p_ul_data[ul_cnt];
	}

	/* To stop the read mode */
	p_efc->EEFC_FCR =
			EEFC_FCR_FKEY_PASSWD | EEFC_FCR_FARG(0) |
			EEFC_FCR_FCMD(cmd_stop);

	/* Wait for the FRDY bit in the Flash Programming Status Register (EEFC_FSR)
	 * rises.
	 */
	do {
		ul_status = p_efc->EEFC_FSR;
	} while ((ul_status & EEFC_FSR_FRDY) != EEFC_FSR_FRDY);

	p_efc->EEFC_FMR &= ~(0x1u << 16);

	return EFC_RC_OK;
}

_NO_INLINE _RAMFUNC uint32_t eefc_run_command_and_wait(Efc *p_efc, uint32_t ul_command, uint32_t ul_argument)
{
	uint32_t result;
	uint32_t irq_flags = 0;

	/* Unique ID commands are not supported. */
	if (ul_command == EEFC_FCR_FCMD_STUI || ul_command == EEFC_FCR_FCMD_SPUI) {
		return EFC_RC_NOT_SUPPORT;
	}
	
	/* Use RAM Function. */
	irq_flags = cpu_irq_save();
		
	volatile uint32_t ul_status;

	p_efc->EEFC_FCR = EEFC_FCR_COMMAND(ul_command, ul_argument);
	do {
		ul_status = p_efc->EEFC_FSR;
	} while ((ul_status & EEFC_FSR_FRDY) != EEFC_FSR_FRDY);
	
	result = (ul_status & EEFC_ERROR_FLAGS);
	cpu_irq_restore(irq_flags);
	return result;
}
