/**
 * \file
 *
 * \brief Pulse Width Modulation (PWM) driver for SAM.
 *
 * Copyright (c) 2011-2015 Atmel Corporation. All rights reserved.
 *
 * \asf_license_start
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 *    Atmel microcontroller product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * \asf_license_stop
 *
 */
/*
 * Support and FAQ: visit <a href="http://www.atmel.com/design-support/">Atmel Support</a>
 */

#include "pwm.h"
#include "../pmc.h"



/** Definitions for PWM synchronous channels update mode */
typedef enum {
	PWM_SYNC_UPDATE_MODE_0 = PWM_SCM_UPDM_MODE0,
	PWM_SYNC_UPDATE_MODE_1 = PWM_SCM_UPDM_MODE1,
	PWM_SYNC_UPDATE_MODE_2 = PWM_SCM_UPDM_MODE2
} pwm_sync_update_mode_t;

/** Definitions for PWM event */
typedef enum {
	PWM_EVENT_PERIOD_END = (0 << 10),      /* The channel counter event occurs at the end of the PWM period. */
	PWM_EVENT_PERIOD_HALF_END = (1 << 10)  /* The channel counter event occurs at the half of the PWM period. */
} pwm_counter_event_t;

/** Definitions of PWM register group */
typedef enum {
	PWM_GROUP_CLOCK = (1 << 0),
	PWM_GROUP_DISABLE = (1 << 1),
	PWM_GROUP_MODE = (1 << 2),
	PWM_GROUP_PERIOD = (1 << 3),
	PWM_GROUP_DEAD_TIME = (1 << 4),
	PWM_GROUP_FAULT = (1 << 5)
} pwm_protect_reg_group_t;

/**
 * \defgroup sam_drivers_pwm_group Pulse Width Modulation (PWM)
 *
 * See \ref sam_pwm_quickstart.
 *
 * Driver for the PWM (Pulse Width Modulation). This driver provides access to the
 * main features of the PWM controller.
 *
 * @{
 */

#ifndef PWM_WPCR_WPKEY_PASSWD
#  define PWM_WPCR_WPKEY_PASSWD 0x50574D00
#endif

#ifndef PWM_WPCR_WPCMD_DISABLE_SW_PROT
#  define PWM_WPCR_WPCMD_DISABLE_SW_PROT (PWM_WPCR_WPCMD(0))
#endif

#ifndef PWM_WPCR_WPCMD_ENABLE_SW_PROT
#  define PWM_WPCR_WPCMD_ENABLE_SW_PROT (PWM_WPCR_WPCMD(1))
#endif

#ifndef PWM_WPCR_WPCMD_ENABLE_HW_PROT
#  define PWM_WPCR_WPCMD_ENABLE_HW_PROT (PWM_WPCR_WPCMD(2))
#endif

#define PWM_CLOCK_DIV_MAX  256
#define PWM_CLOCK_PRE_MAX  11

#define MAX_NUM_CHANNELS 4

uint32_t pwm_channel_init(Pwm *p_pwm, pwm_opt_t *p_channel, uint8_t channel);
void pwm_sync_unlock_update(Pwm *p_pwm);


/**
 * \brief Initialize the PWM source clock (clock A and clock B).
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param clock_config PWM clock configuration.
 *
 * \retval 0 if initialization succeeds, otherwise fails.
 */
uint32_t pwm_init(Pwm *p_pwm, pwm_opt_t *settings, uint8_t num_channels)
{
	if(p_pwm == PWM0){
		pmc_enable_periph_clk(PWM0_IRQn);
	}
	else if(p_pwm == PWM1){
		pmc_enable_periph_clk(PWM1_IRQn);
	}
	
	if(num_channels > MAX_NUM_CHANNELS){
		num_channels = MAX_NUM_CHANNELS;
	}
	
	if(settings->enable_period_start_event){
		p_pwm->PWM_CMP[0].PWM_CMPV = 0;
		p_pwm->PWM_CMP[0].PWM_CMPM = PWM_CMPM_CEN;
		p_pwm->PWM_ELMR[0] |= PWM_ELMR_CSEL0;		
	}
	
	for(uint8_t i = 0; i < num_channels; i++){
		pwm_channel_init(p_pwm, settings, i);
	}
	
	pwm_channel_init(p_pwm, settings, 0);
	
	return 0;
}

bool pwm_is_next_period(Pwm *p_pwm){
	return p_pwm->PWM_ISR1 & (PWM_ISR1_CHID0);
}


/**
 * \brief Initialize one PWM channel.
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param p_channel Configurations of the specified PWM channel.
 *
 * \retval 0 if initialization succeeds, otherwise fails.
 */
uint32_t pwm_channel_init(Pwm *p_pwm, pwm_opt_t *p_channel, uint8_t channel)
{
	uint32_t tmp_reg = 0;
	uint32_t ch_num = channel;

	tmp_reg = PWM_CMR_CPRE(PWM_CMR_CPRE_MCK);// | PWM_CMR_DTHI | PWM_CMR_DTLI;
				
	if(p_channel->center_aligned){
		tmp_reg |= PWM_CMR_CALG;
	}
	
	if(p_channel->active_high_out){
		tmp_reg |= PWM_CMR_CPOL;
	}
	if(p_channel->deadtime_enabled){
		tmp_reg |= PWM_CMR_DTE;
		p_pwm->PWM_CH_NUM[ch_num].PWM_DT =
				PWM_DT_DTL(p_channel->deadtime_pwmh_ticks) | 
				PWM_DT_DTH(p_channel->deadtime_pwmh_ticks);
	}
	
	p_pwm->PWM_CH_NUM[ch_num].PWM_CMR = tmp_reg;

	/* Channel Duty Cycle Register */
	p_pwm->PWM_CH_NUM[ch_num].PWM_CDTY = 0;

	/* Channel Period Register */
	p_pwm->PWM_CH_NUM[ch_num].PWM_CPRD = p_channel->pwm_period_ticks;
		

	/* Output Selection Register */
	p_pwm->PWM_OS = 0;

	/* Output Override Value Register */
	p_pwm->PWM_OOV = 0;

	/* Sync Channels Mode Register */
	p_pwm->PWM_SCM |= (1 << ch_num);
	
	pwm_channel_enable(p_pwm, ch_num);

	return 0;
}


uint32_t pwm_calc_period_ticks(uint32_t pwm_frequency, bool center_aligned){
	uint32_t res =  (pmc_getPeripheralClockValue())/(pwm_frequency);
	
	if(center_aligned){
		res /= 2;
	}
	
	return res;
}

uint16_t pwm_calc_deadtime_ticks(uint32_t deadtime_ns){
	return ((pmc_getPeripheralClockValue()/1000000UL)*deadtime_ns)/(1000UL);
}

/**
 * \brief Change the period of the PWM channel.
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param p_channel Configurations of the specified PWM channel.
 * \param ul_period New period value.
 *
 * \retval 0 if changing succeeds, otherwise fails.
 */
uint32_t pwm_channel_update_period(Pwm *p_pwm, pwm_opt_t *p_channel, uint8_t channel)
{
	p_pwm->PWM_CH_NUM[channel].PWM_CPRDUPD = p_channel->pwm_period_ticks;
	return 0;
}

/**
 * \brief Change the duty cycle of the PWM channel.
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param p_channel Configurations of the specified PWM channel.
 * \param ul_duty New duty cycle value.
 *
 * \retval 0 if changing succeeds, otherwise fails.
 */
uint32_t pwm_channel_update_duty(Pwm *p_pwm, __attribute__((unused)) pwm_opt_t *p_channel, uint32_t ul_duty, uint8_t channel)
{
	p_pwm->PWM_CH_NUM[channel].PWM_CDTYUPD = ul_duty;
	return 0;
}

/**
 * \brief Change dead-time value for PWM outputs.
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param p_channel Configurations of the specified PWM channel.
 * \param us_deadtime_pwmh New dead-time value for PWMH output.
 * \param us_deadtime_pwml New dead-time value for PWML output.
 */
void pwm_channel_update_dead_time(Pwm *p_pwm, pwm_opt_t *p_channel, uint8_t channel)
{
	/* Write channel dead time update register */
	p_pwm->PWM_CH_NUM[channel].PWM_DTUPD =
			PWM_DTUPD_DTLUPD(p_channel->deadtime_pwmh_ticks) |
			PWM_DTUPD_DTHUPD(p_channel->deadtime_pwmh_ticks);
}

void pwm_update_channel_data(Pwm *p_pwm, pwm_opt_t *p_channel, uint8_t channel, uint32_t duty){
	
	pwm_channel_update_dead_time(p_pwm, p_channel, channel);
	pwm_channel_update_duty(p_pwm, p_channel, duty, channel);
	pwm_channel_update_period(p_pwm, p_channel, channel);
	
	pwm_sync_unlock_update(p_pwm);
}

void pwm_enable_override(Pwm *p_pwm, uint8_t mask){
	p_pwm->PWM_OSS |= (mask << 0) | (mask << 16); 
}

void pwm_disable_override(Pwm *p_pwm, uint8_t mask){
	p_pwm->PWM_OSC |= (mask << 0) | (mask << 16);
}

/**
 * \brief Enable the PWM channel.
 *
 * \note The PWM channel should be initialized by pwm_channel_init() before it is enabled.
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param ul_channel PWM channel number to enable.
 */
void pwm_channel_enable(Pwm *p_pwm, uint8_t ul_channel)
{
	p_pwm->PWM_ENA = (1 << ul_channel);
}

/**
 * \brief Disable the PWM channel.
 *
 * \note A disabled PWM channel can be re-initialized using pwm_channel_init().
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param ul_channel PWM channel number to disable.
 */
void pwm_channel_disable(Pwm *p_pwm, uint8_t ul_channel)
{
	p_pwm->PWM_DIS = (1 << ul_channel);
}


/**
 * \brief Check which PWM channel is enabled.
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param p_channel Configurations of the specified PWM channel.
 *
 * \return Bitmask of enabled PWM channel(s).
 */
uint32_t pwm_channel_get_status(Pwm *p_pwm)
{
	return p_pwm->PWM_SR;
}

/**
 * \brief Initialize synchronous channels update mode and period.
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param mode Synchronous channels update mode.
 * \param ul_update_period Time between each update of the synchronous channels.
 *
 * \retval 0 if initialization succeeds, otherwise fails.
 */
uint32_t pwm_sync_init(Pwm *p_pwm, pwm_sync_update_mode_t mode,
		uint32_t ul_update_period)
{
	uint32_t sync_mode = p_pwm->PWM_SCM;

	sync_mode &= ~PWM_SCM_UPDM_Msk;
	sync_mode |= mode;

	p_pwm->PWM_SCM = sync_mode;

	p_pwm->PWM_SCUP = PWM_SCUP_UPR(ul_update_period);

	return 0;
}

/**
 * \brief Unlock the update of synchronous channels.
 *
 * \note After unlock, synchronous channels will be updated at the next PWM period.
 *
 * \param p_pwm Pointer to a PWM instance.
 */
void pwm_sync_unlock_update(Pwm *p_pwm)
{
	p_pwm->PWM_SCUC = PWM_SCUC_UPDULOCK;
}

/**
 * \brief Change the wanted time between each update of the synchronous channels.
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param ul_update_period Time between each update of the synchronous channels.
 */
void pwm_sync_change_period(Pwm *p_pwm, uint32_t ul_update_period)
{
	p_pwm->PWM_SCUPUPD = PWM_SCUPUPD_UPRUPD(ul_update_period);
}

/**
 * \brief Get the value of the synchronization update period counter.
 *
 * \param p_pwm Pointer to a PWM instance.
 *
 * \return Value of the synchronization update Period Counter.
 */
uint32_t pwm_sync_get_period_counter(Pwm *p_pwm)
{
	return PWM_SCUP_UPRCNT(p_pwm->PWM_SCUP);
}

/**
 * \brief Enable PWM write protect.
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param ul_group Bitmask of PWM register group.
 * \param b_sw Boolean of software protect. True for software protect and false for hardware protect.
 */
void pwm_enable_protect(Pwm *p_pwm, uint32_t ul_group, bool b_sw)
{
	uint32_t wp = 0;

	if (b_sw) {
		wp = PWM_WPCR_WPKEY_PASSWD | (ul_group << 2) |
				PWM_WPCR_WPCMD_ENABLE_SW_PROT;
	} else {
		wp = PWM_WPCR_WPKEY_PASSWD | (ul_group << 2) |
				PWM_WPCR_WPCMD_ENABLE_HW_PROT;
	}

	p_pwm->PWM_WPCR = wp;
}

/**
 * \brief Disable PWM write protect.
 *
 * \note Only a hardware reset of PWM controller (handled by PMC) can disable hardware write protect.
 *
 * \param p_pwm Pointer to a PWM instance.
 * \param ul_group Bitmask of PWM register group.
 */
void pwm_disable_protect(Pwm *p_pwm, uint32_t ul_group)
{
	p_pwm->PWM_WPCR = PWM_WPCR_WPKEY_PASSWD
			 | (ul_group << 2) | PWM_WPCR_WPCMD_DISABLE_SW_PROT;
}