/**
 * \file
 *
 * \brief SAM USART in SPI mode driver functions.
 *
 * Copyright (c) 2011-2015 Atmel Corporation. All rights reserved.
 *
 * \asf_license_start
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 *    Atmel microcontroller product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * \asf_license_stop
 *
 */
/*
 * Support and FAQ: visit <a href="http://www.atmel.com/design-support/">Atmel Support</a>
 */

#include <sam.h>
#include <stdlib.h>

#include "../../pio.h"
#include "../../pmc.h"
#include "usart_spi.h"
#include "../Common_driver/usart.h"

void usart_spi_enable(Usart *p_usart);

void usart_spi_init(Usart *p_usart, usart_spi_opt_t *options)
{
	//Enable peripheral clock
	if(p_usart == USART0){
		pmc_enable_periph_clk(ID_USART0);
		pio_setMux(PIOB,  0, C); //MISO
		pio_setMux(PIOB, 13, C); //SCK
		pio_setMux(PIOB,  1, C); //MOSI
	}
	if(p_usart == USART1){
		pmc_enable_periph_clk(ID_USART1);
		//Set output pin muxes
		pio_setMux(PIOA, 21, A); //MISO
		pio_setMux(PIOA, 23, A); //SCK
		pio_setMux(PIOB,  4, D); //MOSI
	}
	if(p_usart == USART2){
		pmc_enable_periph_clk(ID_USART2);
		pio_setMux(PIOD, 15, B); //MISO
		pio_setMux(PIOD, 17, B); //SCK
		pio_setMux(PIOD, 16, B); //MOSI
	}

	static uint32_t ul_reg_val = 0;

	/* Reset the USART and shut down TX and RX. */
	usart_reset(p_usart);

	/* Check whether the input values are legal. */
	if (!options || (options->spi_mode > SPI_MODE_3) || usart_set_spi_master_baudrate(p_usart, options->baudrate, pmc_getPeripheralClockValue())) {
		return ;
	}

	/* Configure the character length bit in MR register. */
	ul_reg_val |= US_MR_CHRL_8_BIT;

	/* Set SPI master mode and channel mode. */
	ul_reg_val |= US_MR_USART_MODE_SPI_MASTER | US_MR_CLKO ;

	switch (options->spi_mode) {
		case SPI_MODE_0:
		ul_reg_val |= US_MR_CPHA;
		ul_reg_val &= ~US_MR_CPOL;
		break;

		case SPI_MODE_1:
		ul_reg_val &= ~US_MR_CPHA;
		ul_reg_val &= ~US_MR_CPOL;
		break;

		case SPI_MODE_2:
		ul_reg_val |= US_MR_CPHA;
		ul_reg_val |= US_MR_CPOL;
		break;

		case SPI_MODE_3:
		ul_reg_val &= ~US_MR_CPHA;
		ul_reg_val |= US_MR_CPOL;
		break;

		default:
		break;
	}

	p_usart->US_MR |= ul_reg_val;

	usart_spi_enable(p_usart);
	
}

void usart_spi_enable(Usart *p_usart)
{
	usart_enable_tx(p_usart);
	usart_enable_rx(p_usart);
}

void usart_spi_disable(Usart *p_usart)
{
	usart_disable_tx(p_usart);
	usart_disable_rx(p_usart);
}

uint32_t usart_spi_write_read_packet(Usart *p_usart, const uint8_t *data, size_t len)
{
	unsigned char* ch_ptr = (unsigned char *) data;
	for(size_t i = 0; i < len; i++) {
		usart_putchar(p_usart, *(ch_ptr + i));
		usart_getchar(p_usart, ch_ptr + i);
	}
	return 0;
}

void usart_spi_select_device(Usart *p_usart)
{
	usart_spi_force_chip_select(p_usart);
}

void usart_spi_deselect_device(Usart *p_usart)
{	
	usart_spi_release_chip_select(p_usart);
}