#include "pmc.h"
#include "Flash/eefc.h"
#define MAX_PERIPH_ID    63

/** Bit mask for peripheral clocks (PCER0) */
#define PMC_MASK_STATUS0        (0xFFFFFFFC)

/** Bit mask for peripheral clocks (PCER1) */
#define PMC_MASK_STATUS1        (0xFFFFFFFF)

/** Loop counter timeout value */
#define PMC_TIMEOUT             (2048)


void pmc_mck_set_prescaler(uint32_t pres);

void pmc_mck_set_division(uint32_t div);

void pmc_mck_set_source(uint32_t source);
uint32_t pmc_switch_mck_to_sclk(uint32_t pres);
uint32_t pmc_switch_mck_to_mainck(uint32_t pres);
uint32_t pmc_switch_mck_to_pllack(uint32_t pres);


uint32_t pmc_switch_mck_to_upllck(uint32_t pres);


//Slow clock (SLCK) oscillator and configuration
void pmc_switch_sclk_to_32kxtal(uint32_t ul_bypass);
uint32_t pmc_osc_is_ready_32kxtal(void);

//Main Clock (MAINCK) oscillator and configuration
void pmc_switch_mainck_to_fastrc(uint32_t ul_moscrcf);
void pmc_osc_enable_fastrc(uint32_t ul_rc);
void pmc_osc_disable_fastrc(void);
uint32_t pmc_osc_is_ready_fastrc(void);
void pmc_osc_enable_main_xtal(uint32_t ul_xtal_startup_time);
void pmc_osc_bypass_main_xtal(void);
void pmc_osc_disable_main_xtal(void);
uint32_t pmc_osc_is_bypassed_main_xtal(void);
uint32_t pmc_osc_is_ready_main_xtal(void);
void pmc_switch_mainck_to_xtal(uint32_t ul_xtal_startup_time);
void pmc_osc_disable_xtal(uint32_t ul_bypass);
uint32_t pmc_osc_is_ready_mainck(void);
void pmc_mainck_osc_select(uint32_t ul_xtal_rc);


//PLL oscillator and configuration
void pmc_enable_pllack(uint32_t mula, uint32_t pllacount, uint32_t diva);
void pmc_disable_pllack(void);
uint32_t pmc_is_locked_pllack(void);



uint32_t pmc_is_locked_upll(void);


//Peripherals clock configuration
void pmc_enable_all_periph_clk(void);
void pmc_disable_all_periph_clk(void);
uint32_t pmc_is_periph_clk_enabled(uint32_t irqnNumber);

//Programmable clock Source and Prescaler configuration
//The following functions may be used to select the clock source and
//prescaler for the specified programmable clock.

void pmc_programmable_clk_set_prescaler(ProgrammableClockSource irqnNumber, uint32_t pres);
void pmc_programmable_clk_set_source(ProgrammableClockSource irqnNumber, uint32_t source);
uint32_t pmc_switch_programmable_clk_to_sclk(ProgrammableClockSource irqnNumber, uint32_t pres);

void pmc_enable_all_programmable_clk(void);
void pmc_disable_all_programmable_clk(void);
uint32_t pmc_is_programmable_clk_enabled(ProgrammableClockSource irqnNumber);
//USB clock configuration

//Interrupt and status management
void pmc_enable_interrupt(uint32_t sources);
void pmc_disable_interrupt(uint32_t sources);
uint32_t pmc_get_interrupt_mask(void);
uint32_t pmc_get_status(void);
void pmc_initPLLAExternal(struct PmcInit pmc_init_struct);
void pmc_enable_upll_clock(void);

static uint32_t pmc_core_clock = 0;

void pmc_init_clocks(struct PmcInit pmc_init_struct){
	
	
	pmc_initPLLAExternal(pmc_init_struct);
	
	pmc_core_clock = pmc_calcCoreClock(pmc_init_struct);
	
	switch(pmc_init_struct.main_freq){
		case EXT_12MHZ:
			// Needed for CAN and USB
			UTMI->UTMI_CKTRIM = 0; // UTMI PLL REFERENCE OSCILLATOR = 12 MHz
			pmc_enable_upll_clock();
		break;
		case EXT_16MHZ:
			// Needed for CAN and USB
			UTMI->UTMI_CKTRIM = 1; // UTMI PLL REFERENCE OSCILLATOR = 16 MHz
			pmc_enable_upll_clock();
		break;
		case INT_4MHZ:
		default:
			//Incorrect input clock for UPLL, CAN and USB will not be available.
		break;
	}
	
}

inline uint32_t pmc_getCoreClockValue(){
	return pmc_core_clock;
}

inline uint32_t pmc_getPeripheralClockValue(){
	//Can't change this setting. Don't know why.
	return pmc_core_clock/2;
}

//Peripheralclock = F_CPU/2

uint32_t pmc_calcCoreClock(struct PmcInit pmc_init_struct){
	
	uint32_t mainosc_frequency = 0;
	
	switch(pmc_init_struct.main_freq){
		case EXT_12MHZ:
			mainosc_frequency = 12000000UL;
		break;
		case EXT_16MHZ:
			mainosc_frequency = 16000000UL;
		break;
		case INT_4MHZ:
			mainosc_frequency =  4000000UL;
		break;
		default:
			mainosc_frequency =			 0;//Incorrect input clock for UPLL, CAN and USB will not be available.
		break;
	}
	
	if(pmc_init_struct.divide == 0){
		return mainosc_frequency;
	}
	else{
		uint32_t clock;
		clock = (mainosc_frequency*pmc_init_struct.multiply)/(pmc_init_struct.divide);
		
		switch(pmc_init_struct.pres){
			case CLK_1:
				clock /= 1;
			break;
			case CLK_2:
				clock /= 2;
			break;
			case CLK_3:
				clock /= 3;
			break;
			case CLK_4:
				clock /= 4;
			break;
			case CLK_8:
				clock /= 8;
			break;
			case CLK_16:
				clock /= 16;
			break;
			case CLK_32:
				clock /= 32;
			break;
			case CLK_64:
				clock /= 64;
			break;
		}
		return clock;
	}
}

void pmc_initPLLAExternal(struct PmcInit pmc_init_struct) {
	// After running this init the peripheral 
	// clock will be half the processor clock frequency.
	
	// Init main oscillator
	pmc_osc_enable_main_xtal(0x8U);
	pmc_switch_mainck_to_xtal(0x8U);
	pmc_osc_disable_fastrc();
	
	//Switch to main clock
	pmc_switch_mck_to_mainck(pmc_init_struct.pres);
	
	// Initialize PLLA
	pmc_enable_pllack((pmc_init_struct.multiply-1), 0x3fU,pmc_init_struct.divide);
	
	//Swich to main clock
	//?
	pmc_switch_mck_to_mainck(pmc_init_struct.pres);
	//Switch to PLLA
	pmc_switch_mck_to_pllack(pmc_init_struct.pres);
	
}

void pmc_enable_writeprotect(void)
{
	PMC->PMC_WPMR = PMC_WPMR_WPKEY_PASSWD | PMC_WPMR_WPEN;
}

void pmc_disable_writeprotect(void)
{
	PMC->PMC_WPMR = PMC_WPMR_WPKEY_PASSWD;
}


void pmc_mck_set_prescaler(uint32_t pres)
{
	pmc_disable_writeprotect();	
	PMC->PMC_MCKR =
			(PMC->PMC_MCKR & (~PMC_MCKR_PRES_Msk)) | pres;
	while (!(PMC->PMC_SR & PMC_SR_MCKRDY));
	pmc_enable_writeprotect();
}


void pmc_mck_set_division(uint32_t div)
{
	pmc_disable_writeprotect();
	switch (div) {
		case 1:
			div = PMC_MCKR_MDIV_EQ_PCK;
			break;
		case 2:
			div = PMC_MCKR_MDIV_PCK_DIV2;
			break;
		case 3:
			div = PMC_MCKR_MDIV_PCK_DIV3;
			break;
		case 4:
			div = PMC_MCKR_MDIV_PCK_DIV4;
			break;
		default:
			div = PMC_MCKR_MDIV_EQ_PCK;
			break;
	}
	PMC->PMC_MCKR = (PMC->PMC_MCKR & (~PMC_MCKR_MDIV_Msk)) | div;
	while (!(PMC->PMC_SR & PMC_SR_MCKRDY));
	pmc_enable_writeprotect();
}


void pmc_mck_set_source(uint32_t source)
{
	pmc_disable_writeprotect();
	PMC->PMC_MCKR = (PMC->PMC_MCKR & (~PMC_MCKR_CSS_Msk)) | source;
	while (!(PMC->PMC_SR & PMC_SR_MCKRDY));
	pmc_enable_writeprotect();
}


uint32_t pmc_switch_mck_to_sclk(uint32_t pres)
{
	uint32_t ul_timeout;
	pmc_disable_writeprotect();
	PMC->PMC_MCKR = (PMC->PMC_MCKR & (~PMC_MCKR_CSS_Msk)) |
			PMC_MCKR_CSS_SLOW_CLK;
	for (ul_timeout = PMC_TIMEOUT; !(PMC->PMC_SR & PMC_SR_MCKRDY);
			--ul_timeout) {
		if (ul_timeout == 0) {
			return 1;
		}
	}

	PMC->PMC_MCKR = (PMC->PMC_MCKR & (~PMC_MCKR_PRES_Msk)) | pres;
	for (ul_timeout = PMC_TIMEOUT; !(PMC->PMC_SR & PMC_SR_MCKRDY);
			--ul_timeout) {
		if (ul_timeout == 0) {
			return 1;
		}
	}
	pmc_enable_writeprotect();
	return 0;
}


uint32_t pmc_switch_mck_to_mainck(uint32_t pres)
{
	uint32_t ul_timeout;
	pmc_disable_writeprotect();
	PMC->PMC_MCKR = (PMC->PMC_MCKR & (~PMC_MCKR_CSS_Msk)) | PMC_MCKR_CSS_MAIN_CLK;
	for (ul_timeout = PMC_TIMEOUT; !(PMC->PMC_SR & PMC_SR_MCKRDY);
			--ul_timeout) {
		if (ul_timeout == 0) {
			return 1;
		}
	}

	PMC->PMC_MCKR = (PMC->PMC_MCKR & (~PMC_MCKR_PRES_Msk)) | pres;
	for (ul_timeout = PMC_TIMEOUT; !(PMC->PMC_SR & PMC_SR_MCKRDY);
			--ul_timeout) {
		if (ul_timeout == 0) {
			return 1;
		}
	}

	pmc_enable_writeprotect();
	return 0;
}

uint32_t pmc_switch_mck_to_pllack(uint32_t pres)
{
	pmc_disable_writeprotect();
	
	PMC->PMC_MCKR = (( (PMC->PMC_MCKR | 1<<8 ) & (~PMC_MCKR_PRES_Msk)) | pres);
	
	while (!(PMC->PMC_SR & PMC_SR_MCKRDY));

	PMC->PMC_MCKR = (( (PMC->PMC_MCKR | 1<<8) & (~PMC_MCKR_CSS_Msk)) | PMC_MCKR_CSS_PLLA_CLK) ;
	// only works when MDIV is set.
	//PMC->PMC_MCKR =  (PMC_MCKR_CSS_PLLA_CLK | 1<<8 ) ;

	while (!(PMC->PMC_SR & PMC_SR_MCKRDY)){		
	}
	

	pmc_enable_writeprotect();
	return 0;
}

uint32_t pmc_switch_mck_to_upllck(uint32_t pres)
{
	uint32_t ul_timeout;
	pmc_disable_writeprotect();
	PMC->PMC_MCKR = (PMC->PMC_MCKR & (~PMC_MCKR_PRES_Msk)) | pres;
	for (ul_timeout = PMC_TIMEOUT; !(PMC->PMC_SR & PMC_SR_MCKRDY);
			--ul_timeout) {
		if (ul_timeout == 0) {
			return 1;
		}
	}

	PMC->PMC_MCKR = (PMC->PMC_MCKR & (~PMC_MCKR_CSS_Msk)) |
			PMC_MCKR_CSS_UPLL_CLK;
	for (ul_timeout = PMC_TIMEOUT; !(PMC->PMC_SR & PMC_SR_MCKRDY);
			--ul_timeout) {
		if (ul_timeout == 0) {
			return 1;
		}
	}
	pmc_enable_writeprotect();
	return 0;
}


uint32_t pmc_osc_is_ready_32kxtal(void)
{
	return ((SUPC->SUPC_SR & SUPC_SR_OSCSEL)
			&& (PMC->PMC_SR & PMC_SR_OSCSELS));
}

void pmc_switch_mainck_to_fastrc(uint32_t ul_moscrcf)
{
	pmc_disable_writeprotect();
	/* Enable Fast RC oscillator but DO NOT switch to RC now */
	PMC->CKGR_MOR |= (CKGR_MOR_KEY_PASSWD | CKGR_MOR_MOSCRCEN);

	/* Wait the Fast RC to stabilize */
	while (!(PMC->PMC_SR & PMC_SR_MOSCRCS));

	/* Change Fast RC oscillator frequency */
	PMC->CKGR_MOR = (PMC->CKGR_MOR & ~CKGR_MOR_MOSCRCF_Msk) |
			CKGR_MOR_KEY_PASSWD | ul_moscrcf;

	/* Wait the Fast RC to stabilize */
	while (!(PMC->PMC_SR & PMC_SR_MOSCRCS));

	/* Switch to Fast RC */
	PMC->CKGR_MOR = (PMC->CKGR_MOR & ~CKGR_MOR_MOSCSEL) |
			CKGR_MOR_KEY_PASSWD;
	pmc_enable_writeprotect();
}


void pmc_osc_enable_fastrc(uint32_t ul_rc)
{
	pmc_disable_writeprotect();
	/* Enable Fast RC oscillator but DO NOT switch to RC */
	PMC->CKGR_MOR |= (CKGR_MOR_KEY_PASSWD | CKGR_MOR_MOSCRCEN);
	/* Wait the Fast RC to stabilize */
	while (!(PMC->PMC_SR & PMC_SR_MOSCRCS));

	/* Change Fast RC oscillator frequency */
	PMC->CKGR_MOR = (PMC->CKGR_MOR & ~CKGR_MOR_MOSCRCF_Msk) |
			CKGR_MOR_KEY_PASSWD | ul_rc;
	/* Wait the Fast RC to stabilize */
	while (!(PMC->PMC_SR & PMC_SR_MOSCRCS));
	pmc_enable_writeprotect();
}


void pmc_osc_disable_fastrc(void)
{
	pmc_disable_writeprotect();
	/* Disable Fast RC oscillator */
	PMC->CKGR_MOR = (PMC->CKGR_MOR & ~CKGR_MOR_MOSCRCEN &
					~CKGR_MOR_MOSCRCF_Msk)
				| CKGR_MOR_KEY_PASSWD;
	pmc_enable_writeprotect();
}


uint32_t pmc_osc_is_ready_fastrc(void)
{
	return (PMC->PMC_SR & PMC_SR_MOSCRCS);
	
}

void pmc_osc_enable_main_xtal(uint32_t ul_xtal_startup_time)
{
	pmc_disable_writeprotect();
	uint32_t mor = PMC->CKGR_MOR;
	mor &= ~(CKGR_MOR_MOSCXTBY|CKGR_MOR_MOSCXTEN);
	mor |= CKGR_MOR_KEY_PASSWD | CKGR_MOR_MOSCXTEN |
			CKGR_MOR_MOSCXTST(ul_xtal_startup_time);
	PMC->CKGR_MOR = mor;
	/* Wait the main Xtal to stabilize */
	while (!(PMC->PMC_SR & PMC_SR_MOSCXTS));
	pmc_enable_writeprotect();
}

uint32_t pmc_osc_is_ready_main_xtal(void)
{
	return (PMC->PMC_SR & PMC_SR_MOSCXTS);
}

void pmc_switch_mainck_to_xtal(uint32_t ul_xtal_startup_time)
{
	pmc_disable_writeprotect();
	/* Enable Main Xtal oscillator */
	PMC->CKGR_MOR = (PMC->CKGR_MOR & ~CKGR_MOR_MOSCXTBY) |
			CKGR_MOR_KEY_PASSWD | CKGR_MOR_MOSCXTEN |
			CKGR_MOR_MOSCXTST(ul_xtal_startup_time);
	/* Wait the Xtal to stabilize */
	while (!(PMC->PMC_SR & PMC_SR_MOSCXTS));

	PMC->CKGR_MOR |= CKGR_MOR_KEY_PASSWD | CKGR_MOR_MOSCSEL;
	pmc_enable_writeprotect();
	
}

uint32_t pmc_osc_is_ready_mainck(void)
{
	return PMC->PMC_SR & PMC_SR_MOSCSELS;
}

void pmc_mainck_osc_select(uint32_t ul_xtal_rc)
{
	pmc_disable_writeprotect();
	uint32_t mor = PMC->CKGR_MOR;
	if (ul_xtal_rc) {
		mor |=  CKGR_MOR_MOSCSEL;
	} else {
		mor &= ~CKGR_MOR_MOSCSEL;
	}
	PMC->CKGR_MOR = CKGR_MOR_KEY_PASSWD | mor;
	pmc_enable_writeprotect();
}

void pmc_enable_pllack(uint32_t mula, uint32_t pllacount, uint32_t diva)
{
	pmc_disable_writeprotect();
	/* first disable the PLL to unlock the lock */
	pmc_disable_pllack();

	pmc_disable_writeprotect();
	PMC->CKGR_PLLAR = CKGR_PLLAR_ONE | CKGR_PLLAR_DIVA(diva) | CKGR_PLLAR_PLLACOUNT(pllacount) | CKGR_PLLAR_MULA(mula);

	while ( !(PMC->PMC_SR & PMC_SR_LOCKA) );
	pmc_enable_writeprotect();
}

void pmc_disable_pllack(void)
{
	pmc_disable_writeprotect();
	PMC->CKGR_PLLAR = CKGR_PLLAR_ONE | CKGR_PLLAR_MULA(0);
	pmc_enable_writeprotect();

}

uint32_t pmc_is_locked_pllack(void)
{
	return (PMC->PMC_SR & PMC_SR_LOCKA);
}

void pmc_enable_upll_clock(void)
{
	pmc_disable_writeprotect();
	PMC->CKGR_UCKR = CKGR_UCKR_UPLLCOUNT(3) | CKGR_UCKR_UPLLEN;

	/* Wait UTMI PLL Lock Status */
	while (!(PMC->PMC_SR & PMC_SR_LOCKU));
	pmc_enable_writeprotect();
}

void pmc_disable_upll_clock(void)
{
	pmc_disable_writeprotect();
	PMC->CKGR_UCKR &= ~CKGR_UCKR_UPLLEN;
	pmc_enable_writeprotect();
}


uint32_t pmc_is_locked_upll(void)
{
	return (PMC->PMC_SR & PMC_SR_LOCKU);
	
}

uint32_t pmc_enable_periph_clk(uint32_t irqnNumber)
{
	if (irqnNumber > MAX_PERIPH_ID) {
		return 1;
	}
	pmc_disable_writeprotect();
	if (irqnNumber < 32) {
		if ((PMC->PMC_PCSR0 & (1u << irqnNumber)) != (1u << irqnNumber)) {
			PMC->PMC_PCER0 = 1 << irqnNumber;
		}

	} else {
		irqnNumber -= 32;
		if ((PMC->PMC_PCSR1 & (1u << irqnNumber)) != (1u << irqnNumber)) {
			PMC->PMC_PCER1 = 1 << irqnNumber;
		}

	}
	pmc_enable_writeprotect();
	return 0;
}


uint32_t pmc_disable_periph_clk(uint32_t irqnNumber)
{
	if (irqnNumber > MAX_PERIPH_ID) {
		return 1;
	}
	pmc_disable_writeprotect();
	if (irqnNumber < 32) {
		if ((PMC->PMC_PCSR0 & (1u << irqnNumber)) == (1u << irqnNumber)) {
			PMC->PMC_PCDR0 = 1 << irqnNumber;
		}

	} else {
		irqnNumber -= 32;
		if ((PMC->PMC_PCSR1 & (1u << irqnNumber)) == (1u << irqnNumber)) {
			PMC->PMC_PCDR1 = 1 << irqnNumber;
		}

	}
	pmc_enable_writeprotect();
	return 0;
}


void pmc_enable_all_periph_clk(void)
{
	pmc_disable_writeprotect();
	PMC->PMC_PCER0 = PMC_MASK_STATUS0;
	while ((PMC->PMC_PCSR0 & PMC_MASK_STATUS0) != PMC_MASK_STATUS0);


	PMC->PMC_PCER1 = PMC_MASK_STATUS1;
	while ((PMC->PMC_PCSR1 & PMC_MASK_STATUS1) != PMC_MASK_STATUS1);
	pmc_enable_writeprotect();

}

void pmc_disable_all_periph_clk(void)
{
	pmc_disable_writeprotect();
	PMC->PMC_PCDR0 = PMC_MASK_STATUS0;
	while ((PMC->PMC_PCSR0 & PMC_MASK_STATUS0) != 0);


	PMC->PMC_PCDR1 = PMC_MASK_STATUS1;
	while ((PMC->PMC_PCSR1 & PMC_MASK_STATUS1) != 0);
	pmc_enable_writeprotect();

}

uint32_t pmc_is_periph_clk_enabled(uint32_t irqnNumber)
{
	if (irqnNumber > MAX_PERIPH_ID) {
		return 0;
	}


	if (irqnNumber < 32) {

		if ((PMC->PMC_PCSR0 & (1u << irqnNumber))) {
			return 1;
		} else {
			return 0;
		}

	} else {
		irqnNumber -= 32;
		if ((PMC->PMC_PCSR1 & (1u << irqnNumber))) {
			return 1;
		} else {
			return 0;
		}
	}
}

void pmc_programmable_clk_set_prescaler(ProgrammableClockSource irqnNumber, uint32_t pres)
{
	pmc_disable_writeprotect();
	PMC->PMC_PCK[irqnNumber] =
			(PMC->PMC_PCK[irqnNumber] & ~PMC_PCK_PRES_Msk) | pres;
	while ((PMC->PMC_SCER & (PMC_SCER_PCK0 << irqnNumber))
			&& !(PMC->PMC_SR & (PMC_SR_PCKRDY0 << irqnNumber)));
	pmc_enable_writeprotect();
}

void pmc_programmable_clk_set_source(ProgrammableClockSource irqnNumber, uint32_t source)
{
	pmc_disable_writeprotect();
	PMC->PMC_PCK[irqnNumber] =
			(PMC->PMC_PCK[irqnNumber] & ~PMC_PCK_CSS_Msk) | source;
	while ((PMC->PMC_SCER & (PMC_SCER_PCK0 << irqnNumber))
			&& !(PMC->PMC_SR & (PMC_SR_PCKRDY0 << irqnNumber)));
	pmc_enable_writeprotect();
}



uint32_t pmc_switch_programmable_clk_to_mainck(ProgrammableClockSource irqnNumber, uint32_t pres)
{
	uint32_t ul_timeout;
	pmc_disable_writeprotect();
	PMC->PMC_PCK[irqnNumber] = PMC_PCK_CSS_MAIN_CLK | ((pres-1) << PMC_PCK_PRES_Pos);
	for (ul_timeout = PMC_TIMEOUT;
	!(PMC->PMC_SR & (PMC_SR_PCKRDY0 << irqnNumber)); --ul_timeout) {
		if (ul_timeout == 0) {
			return 1;
		}
	}
	pmc_enable_writeprotect();
	return 0;
}

uint32_t pmc_switch_programmable_clk_to_pllack(ProgrammableClockSource irqnNumber, uint32_t pres)
{
	uint32_t ul_timeout;
	pmc_disable_writeprotect();
	PMC->PMC_PCK[irqnNumber] = PMC_PCK_CSS_PLLA_CLK | pres;
	for (ul_timeout = PMC_TIMEOUT;
	!(PMC->PMC_SR & (PMC_SR_PCKRDY0 << irqnNumber)); --ul_timeout) {
		if (ul_timeout == 0) {
			return 1;
		}
	}
	pmc_enable_writeprotect();
	return 0;
}

uint32_t pmc_switch_programmable_clk_to_upllck(ProgrammableClockSource irqnNumber, uint32_t pres)
{
	uint32_t ul_timeout;
	pmc_disable_writeprotect();
	PMC->PMC_PCK[irqnNumber] = PMC_PCK_CSS_UPLL_CLK | pres;
	for (ul_timeout = PMC_TIMEOUT;
			!(PMC->PMC_SR & (PMC_SR_PCKRDY0 << irqnNumber));
			--ul_timeout) {
		if (ul_timeout == 0) {
			return 1;
		}
	}
	pmc_enable_writeprotect();
	return 0;
}



/**
 * \brief Enable the specified programmable clock.
 *
 * \param irqnNumber Id of the programmable clock.
 */
void pmc_enable_programmable_clk(ProgrammableClockSource irqnNumber)
{
	pmc_disable_writeprotect();
	PMC->PMC_SCER |= PMC_SCER_PCK0 << irqnNumber;
	pmc_enable_writeprotect();
}

/**
 * \brief Disable the specified programmable clock.
 *
 * \param irqnNumber Id of the programmable clock.
 */
void pmc_disable_programmable_clk(ProgrammableClockSource irqnNumber)
{
	pmc_disable_writeprotect();
	PMC->PMC_SCDR = PMC_SCER_PCK0 << irqnNumber;
	pmc_enable_writeprotect();
}

/**
 * \brief Enable all programmable clocks.
 */
void pmc_enable_all_programmable_clk(void)
{
	pmc_disable_writeprotect();
	PMC->PMC_SCER = PMC_SCER_PCK0 | PMC_SCER_PCK1 | PMC_SCER_PCK2;
	pmc_enable_writeprotect();
}

/**
 * \brief Disable all programmable clocks.
 */
void pmc_disable_all_programmable_clk(void)
{
	pmc_disable_writeprotect();
	PMC->PMC_SCDR = PMC_SCDR_PCK0 | PMC_SCDR_PCK1 | PMC_SCDR_PCK2;
	pmc_enable_writeprotect();
}

/**
 * \brief Check if the specified programmable clock is enabled.
 *
 * \param irqnNumber Id of the programmable clock.
 *
 * \retval 0 Programmable clock is disabled or unknown.
 * \retval 1 Programmable clock is enabled.
 */
uint32_t pmc_is_programmable_clk_enabled(ProgrammableClockSource irqnNumber)
{
	if (irqnNumber > 6) {
		return 0;
	}

	return (PMC->PMC_SCSR & (PMC_SCSR_PCK0 << irqnNumber));
}

// UPLL at 480 MHz needed for High Speed USB. Also for MCAN 480 MHz div by 24, 12 or 6 to achieve 20, 40 or 80 MHz
/**
 * \brief Switch UDP (USB) clock source selection to UPLL clock.
 *
 * \param usbdiv Clock divisor.
 */
void pmc_switch_udpck_to_upllck(uint32_t usbdiv)
{
	pmc_disable_writeprotect();
	PMC->PMC_USB = PMC_USB_USBS | PMC_USB_USBDIV(usbdiv);
	pmc_enable_writeprotect();
}


/**
 * \brief Enable UDP (USB) clock.
 */
void pmc_enable_udpck(void)
{
	pmc_disable_writeprotect();
	PMC->PMC_SCER = PMC_SCER_USBCLK;
	pmc_enable_writeprotect();
}

/**
 * \brief Disable UDP (USB) clock.
 */
void pmc_disable_udpck(void)
{
	pmc_disable_writeprotect();
	PMC->PMC_SCDR = PMC_SCDR_USBCLK;
	pmc_enable_writeprotect();
}


/**
 * \brief Enable PMC interrupts.
 *
 * \param sources Interrupt sources bit map.
 */
void pmc_enable_interrupt(uint32_t sources)
{
	pmc_disable_writeprotect();
	PMC->PMC_IER = sources;
	pmc_enable_writeprotect();
}

/**
 * \brief Disable PMC interrupts.
 *
 * \param sources Interrupt sources bit map.
 */
void pmc_disable_interrupt(uint32_t sources)
{
	pmc_disable_writeprotect();
	PMC->PMC_IDR = sources;
	pmc_enable_writeprotect();
}

/**
 * \brief Get PMC interrupt mask.
 *
 * \return The interrupt mask value.
 */
uint32_t pmc_get_interrupt_mask(void)
{
	return PMC->PMC_IMR;
}

/**
 * \brief Get current status.
 *
 * \return The current PMC status.
 */
uint32_t pmc_get_status(void)
{
	return PMC->PMC_SR;
}

