#include <stdbool.h>
#include <stdlib.h>

#include "spi.h"
#include "pmc.h"
#include "pio.h"


static void spi_setup_mux(struct SpiMasterSettings spi_mux_settings);

static void spi_reset(Spi *spi);
static void spi_enable(Spi *spi);
static void spi_disable(Spi *spi);
static void spi_set_lastxfer(Spi *spi);
static void spi_enable_clock(Spi *spi);
static void spi_setBaudRateHz(uint32_t peripheral_clock_hz, uint32_t baud_rate_hz, enum SpiChipSelect chip_select);

//*******************************************************************************************
void spi_master_init(Spi *spi, struct SpiMasterSettings spi_settings) {
		
	spi_enable_clock(spi);
		
	spi_disable(spi);
	spi_reset(spi);
	spi_set_lastxfer(spi);
	
	// Variable Peripheral Select
	// Chip selects are directly connected to peripheral device
	spi->SPI_MR = SPI_MR_PS | SPI_MR_MSTR | SPI_MR_MODFDIS | SPI_MR_DLYBCS((spi_settings.CSDelay_ns*(pmc_getPeripheralClockValue()/1000000))/1000);
	spi_enable(spi);
	spi_setup_mux(spi_settings);
}

void spi_chip_select_init(Spi *spi, struct SpiSlaveSettings spi_cs_settings) {
	spi_setBaudRateHz(pmc_getPeripheralClockValue(), spi_cs_settings.spi_baudRate_Hz, spi_cs_settings.chip_select);
	
	spi->SPI_CSR[spi_cs_settings.chip_select] |= spi_cs_settings.spi_mode | spi_cs_settings.data_width;
	if(spi_cs_settings.raise_CS_every_transfer){
		spi->SPI_CSR[spi_cs_settings.chip_select] |= SPI_CSR_CSNAAT;
	}	
}

static void spi_setBaudRateHz(uint32_t peripheral_clock_hz, uint32_t baud_rate_hz, enum SpiChipSelect chip_select) {
	
	if ((peripheral_clock_hz/baud_rate_hz > 255) || ((peripheral_clock_hz/baud_rate_hz) < 1)) {
		while(1);
	}
	else {
		SPI0->SPI_CSR[chip_select] |= ((peripheral_clock_hz/baud_rate_hz)  << 8);
	}
}

static void spi_reset(Spi *spi)
{
	spi->SPI_CR = SPI_CR_SWRST;
}

static void spi_enable(Spi *spi)
{
	spi->SPI_CR = SPI_CR_SPIEN;
}


static void spi_disable(Spi *spi)
{
	spi->SPI_CR = SPI_CR_SPIDIS;
}


static void spi_set_lastxfer(Spi *spi)
{
	spi->SPI_CR = SPI_CR_LASTXFER;
}


static void spi_enable_clock(Spi *spi)
{
	if (spi == SPI0) {
		pmc_enable_periph_clk(ID_SPI0);
	}
	else if (spi == SPI1) {
		pmc_enable_periph_clk(ID_SPI1);
	}
}

static void __attribute__((unused)) spi_disable_clock(Spi *spi)
{
	if (spi == SPI0) {
		pmc_disable_periph_clk(ID_SPI0);
	}
	else if (spi == SPI1) {
		pmc_disable_periph_clk(ID_SPI1);
	}
}

static void spi_setup_mux(struct SpiMasterSettings spi_mux_settings) {
	
	// E70 Only has SPI0
	// SPI0 MISO
	pio_setMux(PIOD,20,B);
	// SPI0 MOSI
	pio_setMux(PIOD,21,B);
	// SPI0 SPCK
	pio_setMux(PIOD,22,B);

	//NPCS0 pin
	switch(spi_mux_settings.cs_0){
		case PB2:
		pio_setMux(PIOB, 2, D);
		break;
		default:
		break;
	}
	//NPCS1 pin
	switch(spi_mux_settings.cs_1) {
		case PA31:
		pio_setMux(PIOA, 31, A);
		break;
		case PD25:
		pio_setMux(PIOD, 25, B);
		break;
		default:
		break;
	}
	//NPCS2 pin
	switch(spi_mux_settings.cs_2){
		case PD12:
		pio_setMux(PIOD, 12, C);
		break;
		default:
		break;
	}
	//NPCS3 pin
	switch(spi_mux_settings.cs_3){
		case PD27:
		pio_setMux(PIOD, 27, B);
		break;
		default:
		break;
	}
}

spi_status_t spi_transceive(uint16_t *tx_buffer,  uint16_t *rx_buffer, uint32_t buffer_size, enum SpiChipSelect chip_select) {
	uint32_t timeout = SPI_TIMEOUT;
	bool is_last_transfer;
	uint32_t i = 0;
	for (i = 0; i < buffer_size; i++) {
		
		if (i == (buffer_size-1)) {
			is_last_transfer = true;
		}
		else {
			is_last_transfer = false;
		}
		
		SPI0->SPI_TDR = spi_generateControlWord(is_last_transfer, chip_select, tx_buffer[i]);
		while (!(SPI0->SPI_SR & SPI_SR_TXEMPTY)) {
			if (!timeout--) {
				return SPI_ERROR_TIMEOUT;
			}
		}
		if(rx_buffer != NULL){
			rx_buffer[i] = (SPI0->SPI_RDR & 0xFFFF);
		}
	}
	return SPI_OK;
}

uint32_t spi_generateControlWord(bool last_xfer,enum SpiChipSelect chip_select, uint16_t data) {
	uint32_t chip_select_correct_format = 0;
	switch(chip_select) {
		case NPCS0:
		chip_select_correct_format = 0;
		break;
		case NPCS1:
		chip_select_correct_format = 1 << 16;
		break;
		case NPCS2:
		chip_select_correct_format = 3 << 16;
		break;
		case NPCS3:
		chip_select_correct_format = 7 << 16;
		break;
	}
	if (last_xfer) {
		return ( SPI_TDR_LASTXFER | chip_select_correct_format | data);
	}
	else {
		return (chip_select_correct_format | data);
	}
}

const volatile uint32_t *spi_get_rx_register_address(Spi *spi){
	return &(spi->SPI_RDR);
}
volatile uint32_t *spi_get_tx_register_address(Spi *spi){
	return &(spi->SPI_TDR);
}

uint32_t spi_flush_rx_register(Spi *spi){
	return spi->SPI_RDR;
}