﻿/*
 * uart.c
 *
 * Created: 02.12.2015 18:05:32
 *  Author: Ørjan
 */ 

#include <sam.h>
#include "uart.h"
#include "pmc.h"
#include "pio.h"
#include <stdbool.h>


#define ENABLE_WRITE_PROTECTION (UART_WPMR_WPKEY_PASSWD | UART_WPMR_WPEN)
#define DISABLE_WRITE_PROTECTION UART_WPMR_WPKEY_PASSWD

#define UART_PERIPHERAL_CLOCK UART_MR_BRSRCCK_PERIPH_CLK
#define UART_PERIPHERAL_CLOCK_FREQUENCY pmc_getPeripheralClockValue()
#define BAUDRATE_MINIMUM_DIVISOR 16

static void uart_enableWriteProtection(Uart * uartID)
{
	uartID->UART_WPMR = ENABLE_WRITE_PROTECTION;
}

static void uart_disableWriteProtection(Uart * uartID)
{
	uartID->UART_WPMR = DISABLE_WRITE_PROTECTION;
}

static void uart_setBaudRateClockDivisor(Uart * uartID, uint32_t baudrate)
{  
	uartID->UART_BRGR = (uint16_t)(UART_PERIPHERAL_CLOCK_FREQUENCY/BAUDRATE_MINIMUM_DIVISOR/baudrate);
}

static void uart_setParity(Uart * uartID, enum uart_parity parity)
{
	switch(parity){
		case EVEN_PARITY:
			uartID->UART_MR |= UART_MR_PAR_EVEN;
			break;
		case ODD_PARITY:
			uartID->UART_MR |= UART_MR_PAR_ODD;
			break;
		case SPACE_PARITY:
			uartID->UART_MR |= UART_MR_PAR_SPACE;
			break;
		case MARK_PARITY:
			uartID->UART_MR |= UART_MR_PAR_MARK;
			break;
		case NO_PARITY:
			uartID->UART_MR |= UART_MR_PAR_NO;
			break;	
	}
}

static void uart_setMode(Uart * uartID, enum uart_mode mode)
{
	switch(mode){
		case NORMAL_MODE:
			uartID->UART_MR |= UART_MR_CHMODE_NORMAL;
			break;
		case AUTOMATIC_MODE:
			uartID->UART_MR |= UART_MR_CHMODE_AUTOMATIC;
			break;
		case LOCAL_LOOPBACK_MODE:
			uartID->UART_MR |= UART_MR_CHMODE_LOCAL_LOOPBACK;
			break;
		case REMOTE_LOOPBACK_MODE:
			uartID->UART_MR |= UART_MR_CHMODE_REMOTE_LOOPBACK;
			break;	
	}
}

static void uart_muxSetup(enum uart_TX_pin TX_pin, enum uart_RX_pin RX_pin)
{
	switch(TX_pin)
	{
		case UART0_TX_PIN_PA10:
			pio_setMux(PIOA, 10, A);
			break;
		case UART1_TX_PIN_PA4:
			pio_setMux(PIOA,  4, C);
			break;
		case UART1_TX_PIN_PA6:
			pio_setMux(PIOA,  6, C);
			break;
		case UART1_TX_PIN_PD26:
			pio_setMux(PIOD, 26, D);
			break;
		case UART2_TX_PIN_PD26:
			pio_setMux(PIOD, 26, C);
			break;
		case UART3_TX_PIN_PD30:
			pio_setMux(PIOD, 30, A);
			break;
		case UART3_TX_PIN_PD31:
			pio_setMux(PIOD, 31, B);
			break;
		case UART4_TX_PIN_PD3:
			pio_setMux(PIOD,  3, C);
			break;
		case UART4_TX_PIN_PD19:
			pio_setMux(PIOD, 19, C);
			break;
		default:
			// do nothing
			break;
	}
	
	switch(RX_pin)
	{
		case UART0_RX_PIN_PA9:
			pio_setMux(PIOA,  9, A);
			break;
		case UART1_RX_PIN_PA5:
			pio_setMux(PIOA,  5, C);
			break;
		case UART2_RX_PIN_PD25:
			pio_setMux(PIOD, 25, C);
			break;
		case UART3_RX_PIN_PD28:
			pio_setMux(PIOD, 28, A);
			break;
		case UART4_RX_PIN_PD18:
			pio_setMux(PIOD, 18, C);
			break;
		default:
			// do nothing
			break;
	}
}

static void uart_enableUartClock(Uart * uartID)
{
	if ( uartID == UART0 )
	{
		pmc_enable_periph_clk(UART0_IRQn);
	} 
	else if ( uartID == UART1 )
	{
		pmc_enable_periph_clk(UART1_IRQn);
	}
	else if ( uartID == UART2 )
	{
		pmc_enable_periph_clk(UART2_IRQn);
	}
	else if ( uartID == UART3 )
	{
		pmc_enable_periph_clk(UART3_IRQn);
	}
	else if ( uartID == UART4 )
	{
		pmc_enable_periph_clk(UART4_IRQn);
	}
}

static void uart_setBaudrateSourceClock(Uart * uartID, uint32_t clocksource)
{
	uartID->UART_MR |= clocksource;
}

static void uart_enableUart(Uart * uartID)
{
	uartID->UART_CR = UART_CR_RXEN | UART_CR_TXEN | UART_CR_RSTSTA;
}

void uart_init(struct uart_config config)
{	
	uart_disableWriteProtection(config.uartID);
	
	uart_muxSetup(config.TX_pin, config.RX_pin);
	uart_enableUartClock(config.uartID);
	uart_enableUart(config.uartID);
	uart_setBaudrateSourceClock(config.uartID, UART_PERIPHERAL_CLOCK);
	uart_setBaudRateClockDivisor(config.uartID, config.baudrate);
	uart_setParity(config.uartID, config.parity);
	uart_setMode(config.uartID, config.mode);
	
	uart_enableWriteProtection(config.uartID);
}

enum uart_receiveStatus uart_receiveData(Uart * uartID, uint8_t * data)
{	
	volatile bool new_data_received = (bool)(uartID->UART_SR & UART_SR_RXRDY);
	
	if ( new_data_received )
	{
		*data = uartID->UART_RHR;
		return UART_RECEIVE_SUCCESS;	
	}
	else {
		return UART_RECEIVE_FAILURE;
	}
}

void uart_transmitData(Uart * uartID, uint8_t data)
{	

	//bool transmitter_ready = (bool)(uartID->UART_SR & UART_SR_TXEMPTY);
	uartID->UART_THR = data;
	while ( ! (uartID->UART_SR & UART_SR_TXRDY));
}

void uart_transmitBuffer(Uart *uartID, uint8_t *tx_buffer, uint32_t buffer_size) {
	uint32_t i = 0;
	for (i = 0; i < buffer_size; i++) {
		uart_transmitData(uartID,tx_buffer[i]);
	}
}

void uart_enableCompareInterrupt(Uart *uartID) {
	uartID->UART_IER |= UART_IER_CMP;
}
void uart_enableRxReadyInterrupt(Uart *uartID) {
	uartID->UART_IER |= UART_IER_RXRDY;
}


void uart_setCompareRegisterFlagOnly(Uart *uartID, uint8_t cmp_value_1, uint8_t cmp_value_2) {
	uart_disableWriteProtection(uartID);
	uartID->UART_CMPR = UART_CMPR_VAL1(cmp_value_1) | UART_CMPR_VAL2(cmp_value_2);				
	uart_enableWriteProtection(uartID);
}

void uart_resetStatusRegister(Uart *uartID) {
	uartID->UART_CR = UART_CR_RSTSTA;
}