/*
 * Inverter_Control_16.c
 *
 * Created: 08.02.2016 15.05.37
 * Author : Simen
 */ 


#include <sam.h>
#include <stdlib.h>
#include <arm_math.h>
#include <stdio.h>
#include <math.h>
#include <stdbool.h>
#include "utilities.h"
#include "system_state.h"
#include "Drivers/enDat.h"
#include "Drivers/mazet_status.h"
#include "Sam-base/RevolveDrivers/pio.h"
#include "Sam-base/RevolveDrivers/pmc.h"
#include "Sensors/sensors_current.h"
#include "Sensors/sensors_voltage.h"
#include "Sensors/sensors_encoder.h"
#include "Sensors/sensors_adc.h"
#include "Motor_control/foc.h"
#include "Motor_control/phase_voltage.h"
#include "Motor_control/motor.h"
#include "Motor_control/pid.h"
#include "status.h"
#include "statusCodes.h"
#include "parameter_store.h"
#include "timeBase.h"
#include "errors.h"
#include "can_wrapper.h"

#define TC_TIMER_ALIVE	TIMER0
#define TC_TIMER_STATUS TIMER1
#define TC_TIMER_DATA	TIMER2

state_variables_t state;

int main(void)
{

	NVIC_SetPriorityGrouping(0U);

	WDT->WDT_MR |= 1<<15;
 	
	struct PmcInit pmcInit = {
		.main_freq = EXT_12MHZ,
		.css = PLLA_CLOCK,
		.pres = CLK_1,
		.divide = 1,
		.multiply = 25
	};
		
	system_init_flash(pmc_calcCoreClock(pmcInit));

	pmc_init_clocks(pmcInit);
	
	can_wrapper_init();
	
	pio_init();
	
	parameters_update_and_transmit(&state);

	status_init(TRANSIENT_ERRORS | STATUS_BITS);
	
	phase_voltage_init();
	
	sensors_adc_init();
	
	encoder_init();
	
	state.regulator_data.PID_reg_D.settings.T_s = 1/ 8000.0;
	state.regulator_data.PID_reg_Q.settings.T_s = 1/ 8000.0;
	state.regulator_data.use_mtpa				= true;
	
	state.regulator_data.PID_reg_D.saturation_max = 0.5;
	state.regulator_data.PID_reg_D.saturation_min = -0.5;
	state.regulator_data.PID_reg_Q.saturation_max = 0.5;
	state.regulator_data.PID_reg_Q.saturation_min = -0.5;
	
	tc_initTimer(TC_TIMER_ALIVE, 	1);
	tc_initTimer(TC_TIMER_STATUS,  10);
	tc_initTimer(TC_TIMER_DATA,	  NUM_DATA_MESSAGES * 100);
	pio_setOutput(PIOA, 04, PIN_LOW);
	phase_voltage_is_new_period();
	volatile uint32_t counter = 0;
    while (1) 
    {
		if(phase_voltage_is_new_period()){
			status_signal(LOOP_LOST_SYNC);
		}
		else{
			while(!phase_voltage_is_new_period()){/*counter++;*/}
		}

		state.current_a	 = (-sensors_current_to_A			(sensors_adc_get_value(SENSOR_ANALOG_I_A) ));
		state.current_b	 = (-sensors_current_to_A			(sensors_adc_get_value(SENSOR_ANALOG_I_B) ));
		state.current_dc = (sensors_current_to_A			(sensors_adc_get_value(SENSOR_ANALOG_I_DC)));
		state.voltage_dc = (sensors_voltage_convert_to_V	(sensors_adc_get_value(SENSOR_ANALOG_V_DC)));

		state.encoder_data = enDat_getLastPositionData();
		state.encoder_status = enDat_getLastStatusData();
		
		sensors_adc_start_transfer();
		enDat_startNewTransmission();
				
		check_errors(&state);
		
		if(status_check(PERMANENT_ERRORS | TRANSIENT_ERRORS)){
			state.torque_request = 0;
			state.speed_request = 0;
		}
		
		if(status_check(INVERTER_ENABLED) && !status_check(ENCODER_FAILURE)){
			phase_voltage_enable_phase(PHASE_ALL);
		}
		else{
			phase_voltage_disable_phase(PHASE_ALL);
		}
		
		if(status_check(DERATING_ERRORS)){
			state.torque_request /= 2;
		}
		
		encoder_get_sin_cos(&state);
		foc_do_dq_transform(&state);
		
		foc_calc_setpoints(&state);
		
		state.voltage_d = do_pid(&(state.regulator_data.PID_reg_D), state.setpoint_d - state.current_d );
		state.voltage_q = do_pid(&(state.regulator_data.PID_reg_Q), state.setpoint_q - state.current_q );
		
		
		foc_limit_vector_dq(&(state.setpoint_d), &(state.setpoint_q), 0.5);
		
		foc_do_inverse_dq_transform(&state);
		
		phase_voltage_set_fraction(PHASE_A, state.voltage_a);
		phase_voltage_set_fraction(PHASE_B, state.voltage_b);
		phase_voltage_set_fraction(PHASE_C, state.voltage_a - state.voltage_b);
				
		static bool resend_status_message = false;
		static bool resend_alive_message = false;
		static bool send_data_message = false;
		if(tc_overflowOccurred(TC_TIMER_DATA)){
			if(status_check(READY_TO_DRIVE)){
				state.min_torque = can_wrapper_get_min_torque_request();
				state.max_torque = can_wrapper_get_max_torque_request();
				state.speed_request = can_wrapper_get_speed_request();
				if(status_check(OVERRIDE_SPEED_CONTROLLER)){
					state.torque_request = state.max_torque;
				}
				else{
					state.torque_request = do_pid(&(state.regulator_data.PID_reg_speed), state.speed_request - state.speed );
				}
			}
			else{
				state.torque_request = 0.0;
			}
			encoder_calc_speed(&state, NUM_DATA_MESSAGES * 100);
			//Send state data @100hz
			send_data_message = true;
		}
		if(send_data_message){
			send_data_message = !can_wrapper_send_data_msg(&state);
		}
		if(tc_overflowOccurred(TC_TIMER_STATUS) || resend_status_message){
			//Send status message @10hz
			resend_status_message = !can_wrapper_send_state_msg();
			counter = 0;
			status_clear(LOOP_LOST_SYNC);
		}

		if(tc_overflowOccurred(TC_TIMER_ALIVE) || resend_alive_message){
			//Send alive message @1hz
			//pio_toggleOutput(PIOA, 21);
			resend_alive_message = !can_wrapper_send_alive_msg();
		}
    }
}